MAX_WIDTH = 26

text_words = {
    "Attack ": b"\xD6\x00",
    "Angel ": b"\xD6\x01",
    "After ": b"\xD6\x02",
    "Aura ": b"\xD6\x03",
    "Ankor ": b"\xD6\x04",
    "Black ": b"\xD6\x05",
    "Bill: ": b"\xD6\x06",
    "Crystal ": b"\xD6\x07",
    "City ": b"\xD6\x08",
    "Come ": b"\xD6\x09",
    "Condor's ": b"\xD6\x0A",
    "Change ": b"\xD6\x0B",
    "Dark ": b"\xD6\x0C",
    "Don't ": b"\xD6\x0D",
    "Diamond ": b"\xD6\x0E",
    "Drifting, ": b"\xD6\x0F",
    "Eric: ": b"\xD6\x10",
    "Edward ": b"\xD6\x11",
    "Even ": b"\xD6\x12",
    "Elder2f ": b"\xD6\x13",
    "Earth ": b"\xD6\x14",
    "Freedan's ": b"\xD6\x15",
    "Great ": b"\xD6\x16",
    "Grandma ": b"\xD6\x17",
    "Good. ": b"\xD6\x18",
    "Gold ": b"\xD6\x19",
    "Grandpa ": b"\xD6\x1A",
    "Hieroglyph ": b"\xD6\x1B",
    "Hey, ": b"\xD6\x1C",
    "It's ": b"\xD6\x1D",
    "Inca ": b"\xD6\x1E",
    "I'll ": b"\xD6\x1F",
    "I've ": b"\xD6\x20",
    "Itorie ": b"\xD6\x21",
    "Jewels ": b"\xD6\x22",
    "Jewels! ": b"\xD6\x23",
    "Just ": b"\xD6\x24",
    "Kara: ": b"\xD6\x25",
    "Kara ": b"\xD6\x26",
    "King ": b"\xD6\x27",
    "Knight ": b"\xD6\x28",
    "Karen's ": b"\xD6\x29",
    "Lilly: ": b"\xD6\x2A",
    "Let's ": b"\xD6\x2B",
    "Lilly ": b"\xD6\x2C",
    "Lola's ": b"\xD6\x2D",
    "Lola: ": b"\xD6\x2E",
    "Mystery ": b"\xD6\x2F",
    "Maybe ": b"\xD6\x30",
    "Moon ": b"\xD6\x31",
    "Man: ": b"\xD6\x32",
    "Morris's ": b"\xD6\x33",
    "Melody ": b"\xD6\x34",
    "Morris: ": b"\xD6\x35",
    "Neil: ": b"\xD6\x36",
    "Neil's ": b"\xD6\x37",
    "Only ": b"\xD6\x38",
    "Once ": b"\xD6\x39",
    "Oink ": b"\xD6\x3A",
    "Please ": b"\xD6\x3B",
    "Psycho ": b"\xD6\x3C",
    "People ": b"\xD6\x3D",
    "Prayer ": b"\xD6\x3E",
    "Pyramid ": b"\xD6\x3F",
    "Purification ": b"\xD6\x40",
    "Plate ": b"\xD6\x41",
    "Quit ": b"\xD6\x42",
    "Rob: ": b"\xD6\x43",
    "Rolek ": b"\xD6\x44",
    "Soldier: ": b"\xD6\x45",
    "Strange ": b"\xD6\x46",
    "South ": b"\xD6\x47",
    "Statue ": b"\xD6\x48",
    "Somehow ": b"\xD6\x49",
    "Sometimes ": b"\xD6\x4A",
    "Something ": b"\xD6\x4B",
    "Shadow's ": b"\xD6\x4C",
    "Someone ": b"\xD6\x4D",
    "This ": b"\xD6\x4E",
    "Will: ": b"\xD6\x4F",
    "There's ": b"\xD6\x50",
    "Will's ": b"\xD6\x51",
    "There ": b"\xD6\x52",
    "That's ": b"\xD6\x53",
    "Tower ": b"\xD6\x54",
    "They ": b"\xD6\x55",
    "Then ": b"\xD6\x56",
    "Trip ": b"\xD6\x57",
    "Thank ": b"\xD6\x58",
    "Take ": b"\xD6\x59",
    "Will. ": b"\xD6\x5A",
    "That ": b"\xD6\x5B",
    "Will, ": b"\xD6\x5C",
    "They're ": b"\xD6\x5D",
    "These ": b"\xD6\x5E",
    "Vampire ": b"\xD6\x5F",
    "Village. ": b"\xD6\x60",
    "When ": b"\xD6\x61",
    "What ": b"\xD6\x62",
    "Well, ": b"\xD6\x63",
    "What's ": b"\xD6\x64",
    "Where ": b"\xD6\x65",
    "Woman: ": b"\xD6\x66",
    "You've ": b"\xD6\x67",
    "Your ": b"\xD6\x68",
    "You're ": b"\xD6\x69",
    "Yes, ": b"\xD6\x6A",
    "about ": b"\xD6\x6B",
    "anything ": b"\xD6\x6C",
    "around ": b"\xD6\x6D",
    "another ": b"\xD6\x6E",
    "ancient ": b"\xD6\x6F",
    "been ": b"\xD6\x70",
    "become ": b"\xD6\x71",
    "body ": b"\xD6\x72",
    "back ": b"\xD6\x73",
    "before ": b"\xD6\x74",
    "brought ": b"\xD6\x75",
    "beautiful ": b"\xD6\x76",
    "being ": b"\xD6\x77",
    "can't ": b"\xD6\x78",
    "come ": b"\xD6\x79",
    "could ": b"\xD6\x7A",
    "comet ": b"\xD6\x7B",
    "company ": b"\xD6\x7C",
    "children ": b"\xD6\x7D",
    "constellation ": b"\xD6\x7E",
    "changed ": b"\xD6\x7F",
    "came ": b"\xD6\x80",
    "coming ": b"\xD6\x81",
    "don't ": b"\xD6\x82",
    "didn't ": b"\xD6\x83",
    "doesn't ": b"\xD6\x84",
    "distant ": b"\xD6\x85",
    "different ": b"\xD6\x86",
    "demons ": b"\xD6\x87",
    "destroy ": b"\xD6\x88",
    "everyone ": b"\xD6\x89",
    "explorer ": b"\xD6\x8A",
    "everyone's ": b"\xD6\x8B",
    "enemies ": b"\xD6\x8C",
    "exposed ": b"\xD6\x8D",
    "from ": b"\xD6\x8E",
    "found ": b"\xD6\x8F",
    "find ": b"\xD6\x90",
    "feel ": b"\xD6\x91",
    "father's ": b"\xD6\x92",
    "going ": b"\xD6\x93",
    "good ": b"\xD6\x94",
    "great ": b"\xD6\x95",
    "ground ": b"\xD6\x96",
    "give ": b"\xD6\x97",
    "have ": b"\xD6\x98",
    "heard ": b"\xD6\x99",
    "human ": b"\xD6\x9A",
    "hear ": b"\xD6\x9B",
    "huge ": b"\xD6\x9C",
    "happened ": b"\xD6\x9D",
    "hieroglyph ": b"\xD6\x9E",
    "it's ": b"\xD6\x9F",
    "inventory ": b"\xD6\xA0",
    "into ": b"\xD6\xA1",
    "inside ": b"\xD6\xA2",
    "just ": b"\xD6\xA3",
    "know ": b"\xD6\xA4",
    "like ": b"\xD6\xA5",
    "long ": b"\xD6\xA6",
    "little ": b"\xD6\xA7",
    "light ": b"\xD6\xA8",
    "look ": b"\xD6\xA9",
    "looks ": b"\xD6\xAA",
    "looking ": b"\xD6\xAB",
    "leave ": b"\xD6\xAC",
    "labor ": b"\xD6\xAD",
    "left ": b"\xD6\xAE",
    "live ": b"\xD6\xAF",
    "life ": b"\xD6\xB0",
    "living ": b"\xD6\xB1",
    "must ": b"\xD6\xB2",
    "made ": b"\xD6\xB3",
    "melody ": b"\xD6\xB4",
    "move ": b"\xD6\xB5",
    "many ": b"\xD6\xB6",
    "more ": b"\xD6\xB7",
    "matter ": b"\xD6\xB8",
    "nothing ": b"\xD6\xB9",
    "need ": b"\xD6\xBA",
    "never ": b"\xD6\xBB",
    "next ": b"\xD6\xBC",
    "other ": b"\xD6\xBD",
    "over ": b"\xD6\xBE",
    "outside ": b"\xD6\xBF",
    "original ": b"\xD6\xC0",
    "people ": b"\xD6\xC1",
    "power ": b"\xD6\xC2",
    "present. ": b"\xD6\xC3",
    "playing ": b"\xD6\xC4",
    "raised ": b"\xD6\xC5",
    "right-hand ": b"\xD6\xC6",
    "strange ": b"\xD6\xC7",
    "something ": b"\xD6\xC8",
    "statue ": b"\xD6\xC9",
    "should ": b"\xD6\xCA",
    "started ": b"\xD6\xCB",
    "seems ": b"\xD6\xCC",
    "same ": b"\xD6\xCD",
    "such ": b"\xD6\xCE",
    "someone ": b"\xD6\xCF",
    "some ": b"\xD6\xD0",
    "save ": b"\xD6\xD1",
    "statues ": b"\xD6\xD2",
    "still ": b"\xD6\xD3",
    "said ": b"\xD6\xD4",
    "stands ": b"\xD6\xD5",
    "this ": b"\xD6\xD6",
    "that ": b"\xD6\xD7",
    "thought ": b"\xD6\xD8",
    "there ": b"\xD6\xD9",
    "think ": b"\xD6\xDA",
    "there's ": b"\xD6\xDB",
    "through ": b"\xD6\xDC",
    "tried ": b"\xD6\xDD",
    "terrible ": b"\xD6\xDE",
    "time ": b"\xD6\xDF",
    "things ": b"\xD6\xE0",
    "their ": b"\xD6\xE1",
    "town ": b"\xD6\xE2",
    "thing ": b"\xD6\xE3",
    "these ": b"\xD6\xE4",
    "temple ": b"\xD6\xE5",
    "them ": b"\xD6\xE6",
    "take ": b"\xD6\xE7",
    "turned, ": b"\xD6\xE8",
    "understand ": b"\xD6\xE9",
    "under ": b"\xD6\xEA",
    "underwater ": b"\xD6\xEB",
    "village ": b"\xD6\xEC",
    "very ": b"\xD6\xED",
    "voice ": b"\xD6\xEE",
    "will ": b"\xD6\xEF",
    "with ": b"\xD6\xF0",
    "want ": b"\xD6\xF1",
    "were ": b"\xD6\xF2",
    "would ": b"\xD6\xF3",
    "where ": b"\xD6\xF4",
    "world ": b"\xD6\xF5",
    "when ": b"\xD6\xF6",
    "what ": b"\xD6\xF7",
    "without ": b"\xD6\xF8",
    "wonder ": b"\xD6\xF9",
    "won't ": b"\xD6\xFA",
    "wouldn't ": b"\xD6\xFB",
    "wanted ": b"\xD6\xFC",
    "waiting ": b"\xD6\xFD",
    "your ": b"\xD6\xFE",
    "you're ": b"\xD6\xFF",
    "1994 ": b"\xD7\x00",
    "Actually, ": b"\xD7\x01",
    "Button ": b"\xD7\x02",
    "Buttons ": b"\xD7\x03",
    "Back ": b"\xD7\x04",
    "Child: ": b"\xD7\x05",
    "Defeating ": b"\xD7\x06",
    "Didn't ": b"\xD7\x07",
    "Desert ": b"\xD7\x08",
    "Erasquez: ": b"\xD7\x09",
    "Elder ": b"\xD7\x0A",
    "Earth's ": b"\xD7\x0B",
    "Eric's ": b"\xD7\x0C",
    "Everybody ": b"\xD7\x0D",
    "Flute: ": b"\xD7\x0E",
    "First ": b"\xD7\x0F",
    "Firebird, ": b"\xD7\x10",
    "From ": b"\xD7\x11",
    "Gaia, ": b"\xD7\x12",
    "Gorgon ": b"\xD7\x13",
    "Have ": b"\xD7\x14",
    "Hey! ": b"\xD7\x15",
    "Istar's ": b"\xD7\x16",
    "Ishtar's ": b"\xD7\x17",
    "Lilly's ": b"\xD7\x18",
    "Larai ": b"\xD7\x19",
    "Looking ": b"\xD7\x1A",
    "Main ": b"\xD7\x1B",
    "Man's ": b"\xD7\x1C",
    "Memory ": b"\xD7\x1D",
    "Mountain ": b"\xD7\x1E",
    "Morris ": b"\xD7\x1F",
    "Many ": b"\xD7\x20",
    "Native ": b"\xD7\x21",
    "Native's ": b"\xD7\x22",
    "Neil ": b"\xD7\x23",
    "Naska ": b"\xD7\x24",
    "Power ": b"\xD7\x25",
    "Prison ": b"\xD7\x26",
    "Play ": b"\xD7\x27",
    "Panther's ": b"\xD7\x28",
    "Rob's ": b"\xD7\x29",
    "Recently ": b"\xD7\x2A",
    "Restores ": b"\xD7\x2B",
    "Right ": b"\xD7\x2C",
    "Russian ": b"\xD7\x2D",
    "Rofsky ": b"\xD7\x2E",
    "Rearrange ": b"\xD7\x2F",
    "Special ": b"\xD7\x30",
    "Spin ": b"\xD7\x31",
    "Seaside ": b"\xD7\x32",
    "She's ": b"\xD7\x33",
    "Shall ": b"\xD7\x34",
    "Sorry ": b"\xD7\x35",
    "Select ": b"\xD7\x36",
    "Show ": b"\xD7\x37",
    "Soon ": b"\xD7\x38",
    "Will! ": b"\xD7\x39",
    "Will... ": b"\xD7\x3A",
    "Tell ": b"\xD7\x3B",
    "Uses ": b"\xD7\x3C",
    "Wind ": b"\xD7\x3D",
    "We'll ": b"\xD7\x3E",
    "We're ": b"\xD7\x3F",
    "With ": b"\xD7\x40",
    "We've ": b"\xD7\x41",
    "Wait ": b"\xD7\x42",
    "Watermia. ": b"\xD7\x43",
    "always ": b"\xD7\x44",
    "approaching ": b"\xD7\x45",
    "animals ": b"\xD7\x46",
    "almost ": b"\xD7\x47",
    "also ": b"\xD7\x48",
    "anything. ": b"\xD7\x49",
    "abolition ": b"\xD7\x4A",
    "breath ": b"\xD7\x4B",
    "birthday ": b"\xD7\x4C",
    "best ": b"\xD7\x4D",
    "bring ": b"\xD7\x4E",
    "bouquet ": b"\xD7\x4F",
    "better ": b"\xD7\x50",
    "behind ": b"\xD7\x51",
    "change ": b"\xD7\x52",
    "castle ": b"\xD7\x53",
    "called ": b"\xD7\x54",
    "comet's ": b"\xD7\x55",
    "condition. ": b"\xD7\x56",
    "care ": b"\xD7\x57",
    "door ": b"\xD7\x58",
    "destroyed ": b"\xD7\x59",
    "disappeared ": b"\xD7\x5A",
    "discovered ": b"\xD7\x5B",
    "dark ": b"\xD7\x5C",
    "ever ": b"\xD7\x5D",
    "elevator ": b"\xD7\x5E",
    "explorer. ": b"\xD7\x5F",
    "eyes ": b"\xD7\x60",
    "first ": b"\xD7\x61",
    "fish ": b"\xD7\x62",
    "followed ": b"\xD7\x63",
    "fountain ": b"\xD7\x64",
    "floor ": b"\xD7\x65",
    "finally ": b"\xD7\x66",
    "father ": b"\xD7\x67",
    "flower ": b"\xD7\x68",
    "forced ": b"\xD7\x69",
    "forget ": b"\xD7\x6A",
    "fate ": b"\xD7\x6B",
    "girl ": b"\xD7\x6C",
    "gets ": b"\xD7\x6D",
    "guess ": b"\xD7\x6E",
    "girl's ": b"\xD7\x6F",
    "house ": b"\xD7\x70",
    "hope ": b"\xD7\x71",
    "home ": b"\xD7\x72",
    "here ": b"\xD7\x73",
    "here. ": b"\xD7\x74",
    "home! ": b"\xD7\x75",
    "happen.... ": b"\xD7\x76",
    "houses ": b"\xD7\x77",
    "inventing ": b"\xD7\x78",
    "last ": b"\xD7\x79",
    "lost ": b"\xD7\x7A",
    "language. ": b"\xD7\x7B",
    "list ": b"\xD7\x7C",
    "life. ": b"\xD7\x7D",
    "laborer ": b"\xD7\x7E",
    "letter ": b"\xD7\x7F",
    "looked ": b"\xD7\x80",
    "lose ": b"\xD7\x81",
    "left-hand ": b"\xD7\x82",
    "mushrooms ": b"\xD7\x83",
    "make ": b"\xD7\x84",
    "mother ": b"\xD7\x85",
    "merchants ": b"\xD7\x86",
    "meet ": b"\xD7\x87",
    "most ": b"\xD7\x88",
    "only ": b"\xD7\x89",
    "ocean ": b"\xD7\x8A",
    "opened ": b"\xD7\x8B",
    "outskirts ": b"\xD7\x8C",
    "president ": b"\xD7\x8D",
    "please ": b"\xD7\x8E",
    "probably ": b"\xD7\x8F",
    "place ": b"\xD7\x90",
    "prison ": b"\xD7\x91",
    "pretty ": b"\xD7\x92",
    "palace ": b"\xD7\x93",
    "right ": b"\xD7\x94",
    "really ": b"\xD7\x95",
    "returned ": b"\xD7\x96",
    "running ": b"\xD7\x97",
    "ruins ": b"\xD7\x98",
    "right. ": b"\xD7\x99",
    "ruins, ": b"\xD7\x9A",
    "road ": b"\xD7\x9B",
    "rest ": b"\xD7\x9C",
    "stone ": b"\xD7\x9D",
    "seem ": b"\xD7\x9E",
    "scattered ": b"\xD7\x9F",
    "seemed ": b"\xD7\xA0",
    "softly ": b"\xD7\xA1",
    "soldiers ": b"\xD7\xA2",
    "shape ": b"\xD7\xA3",
    "since ": b"\xD7\xA4",
    "surprised ": b"\xD7\xA5",
    "sure ": b"\xD7\xA6",
    "seen ": b"\xD7\xA7",
    "shouldn't ": b"\xD7\xA8",
    "somewhere ": b"\xD7\xA9",
    "times ": b"\xD7\xAA",
    "they ": b"\xD7\xAB",
    "talk ": b"\xD7\xAC",
    "tell ": b"\xD7\xAD",
    "townspeople ": b"\xD7\xAE",
    "taken ": b"\xD7\xAF",
    "they're ": b"\xD7\xB0",
    "travelling ": b"\xD7\xB1",
    "trying ": b"\xD7\xB2",
    "turned ": b"\xD7\xB3",
    "temporary ": b"\xD7\xB4",
    "than ": b"\xD7\xB5",
    "then ": b"\xD7\xB6",
    "too, ": b"\xD7\xB7",
    "thousands ": b"\xD7\xB8",
    "turn ": b"\xD7\xB9",
    "treasure, ": b"\xD7\xBA",
    "used ": b"\xD7\xBB",
    "until ": b"\xD7\xBC",
    "village. ": b"\xD7\xBD",
    "vampire ": b"\xD7\xBE",
    "while ": b"\xD7\xBF",
    "water ": b"\xD7\xC0",
    "went ": b"\xD7\xC1",
    "wondered ": b"\xD7\xC2",
    "written ": b"\xD7\xC3",
    "woman ": b"\xD7\xC4",
    "wind ": b"\xD7\xC5",
    "years ": b"\xD7\xC6",
    "you. ": b"\xD7\xC7"
}

text_letters = {
    "?": b"\x0D",
    "'": b"\x0E",
    "0": b"\x20",
    "1": b"\x21",
    "2": b"\x22",
    "3": b"\x23",
    "4": b"\x24",
    "5": b"\x25",
    "6": b"\x26",
    "7": b"\x27",
    "8": b"\x28",
    "9": b"\x29",
    ".": b"\x2A",
    ",": b"\x2B",
    ">": b"\x2C",
    #    '"' : 	b"\x2D",
    #    '"' : 	b"\x2E",
    ":": b"\x2F",
    "A": b"\x40",
    "B": b"\x41",
    "C": b"\x42",
    "D": b"\x43",
    "E": b"\x44",
    "F": b"\x45",
    "G": b"\x46",
    "H": b"\x47",
    "I": b"\x48",
    "J": b"\x49",
    "K": b"\x4A",
    "L": b"\x4B",
    "M": b"\x4C",
    "N": b"\x4D",
    "O": b"\x4E",
    "!": b"\x4F",
    "P": b"\x60",
    "Q": b"\x61",
    "R": b"\x62",
    "S": b"\x63",
    "T": b"\x64",
    "U": b"\x65",
    "V": b"\x66",
    "W": b"\x67",
    "X": b"\x68",
    "Y": b"\x69",
    "Z": b"\x6A",
    "/": b"\x6B",
    "*": b"\x6C",
    "-": b"\x6D",
    "(": b"\x6E",
    ")": b"\x6F",
    "a": b"\x80",
    "b": b"\x81",
    "c": b"\x82",
    "d": b"\x83",
    "e": b"\x84",
    "f": b"\x85",
    "g": b"\x86",
    "h": b"\x87",
    "i": b"\x88",
    "j": b"\x89",
    "k": b"\x8A",
    "l": b"\x8B",
    "m": b"\x8C",
    "n": b"\x8D",
    "o": b"\x8E",
    "p": b"\xA0",
    "q": b"\xA1",
    "r": b"\xA2",
    "s": b"\xA3",
    "t": b"\xA4",
    "u": b"\xA5",
    "v": b"\xA6",
    "w": b"\xA7",
    "x": b"\xA8",
    "y": b"\xA9",
    "z": b"\xAA",
    ",": b"\xAB",
    " ": b"\xAC",
    "|": b"\xCF"  # Prompts for advance
}

text_words_inv = {v: k for k, v in text_words.items()}
text_letters_inv = {v: k for k, v in text_letters.items()}


def get_text(addr, f, rom_offset=0):
    str_encoded = ""
    f.seek(addr + rom_offset)
    char = ""
    bytes = 0
    while char != b"\xC0" and char != b"\xCA":
        char = f.read(1)
        str_encoded += char
        bytes += 1

    str_decoded = ""
    i = 0
    while i < len(str_encoded):
        byte = str_encoded[i]
        i += 1
        if byte not in [b"\xD3", b"\xC0", b"\xCA"]:
            if byte in [b"\xD6", b"\xD7"]:
                word = byte + str_encoded[i]
                i += 1
                if word in text_words_inv:
                    str_decoded += text_words_inv[word]
                # else:
                # print "WARNING: Word not recognized >>", word
            elif byte in text_letters_inv:
                str_decoded += text_letters_inv[byte]
            elif byte not in [b"\xcb"]:
                str_decoded += "?"
                # print "WARNING: Letter not recognized >>", byte

    # print str_decoded
    return [str_decoded, bytes]


def encode(unencoded_str, full_box=False):
    words = []
    word = ""
    i = 0
    while i < len(unencoded_str):
        char = unencoded_str[i]
        if char == "|":
            words.append(word)
            words.append(char)
            word = ""
        else:
            word += char
            if char == " " or i == len(unencoded_str) - 1:
                words.append(word)
                word = ""
        i += 1

    # print words
    if full_box:
        str_encoded = b"\xd3"
    else:
        str_encoded = b""
    text_width = 0

    for word in words:
        if len(word) > MAX_WIDTH:
            print("ERROR: Word too long >>", word)
            return ""
        if word == "|":
            text_width = 0
        else:
            text_width += len(word)
        if text_width > MAX_WIDTH:
            str_encoded += b"\xCB"
            text_width = len(word)
        if word in text_words:
            str_encoded += text_words[word]
        else:
            i = 0
            while i < len(word):
                char = word[i]
                if char in text_letters:
                    str_encoded += text_letters[char]
                else:
                    str_encoded += b"\x0D"
                    print("WARNING: Character not recognized >>", char)
                i += 1

    if full_box:
        str_encoded += b"\xc0"

    # print str_encoded, len(str_encoded)
    return str_encoded



import copy
import hashlib
import json
import logging
import os
import random
import sys

from . import asar
from .classes import World
from .errors import OffsetError
from .models.enums import *
from .models.randomizer_data import RandomizerData

VERSION = "5.1.5.1"
MAX_RANDO_RETRIES = 50
OUTPUT_FOLDER: str = os.path.dirname(os.path.realpath(
    __file__)) + os.path.sep + ".." + os.path.sep + ".." + os.path.sep + "data" + os.path.sep + "output" + os.path.sep


def generate_filename(settings: RandomizerData, extension: str):
    def getDifficulty(difficulty):
        if difficulty.value == Difficulty.EASY.value:
            return "_E"
        if difficulty.value == Difficulty.NORMAL.value:
            return "_N"
        if difficulty.value == Difficulty.HARD.value:
            return "_H"
        if difficulty.value == Difficulty.EXTREME.value:
            return "_X"

    def getGoal(goal, statues, statue_req):
        if goal.value is Goal.DARK_GAIA.value:
            return "D" + statues[0] + getStatueReq(statue_req)
        if goal.value is Goal.APO_GAIA.value:
            return "A" + statues[0] + getStatueReq(statue_req)
        if goal.value is Goal.RANDOM_GAIA.value:
            return "R" + statues[0] + getStatueReq(statue_req)
        if goal.value is Goal.RED_JEWEL_HUNT.value:
            return "J"

    def getStatueReq(statue_req):
        if statue_req.value == StatueReq.PLAYER_CHOICE.value:
            return "P"
        if statue_req.value == StatueReq.RANDOM_CHOICE.value:
            return "R"
        else:
            return "G"

    def getLogic(logic):
        if logic.value == Logic.COMPLETABLE.value:
            return "C"
        if logic.value == Logic.BEATABLE.value:
            return "B"
        if logic.value == Logic.CHAOS.value:
            return "X"

    def getEntranceShuffle(coupled_exits, town_shuffle, dungeon_shuffle, overworld_shuffle):
        if not town_shuffle and not overworld_shuffle and not dungeon_shuffle:
            return ""
        affix = "_er"
        if coupled_exits:
            affix += "C"
        else:
            affix += "U"
        affix += "-"
        if overworld_shuffle:
            affix += "W"
        if town_shuffle:
            affix += "T"
        if dungeon_shuffle:
            affix += "D"
        return affix

    def getDarkRooms(darkrooms):
        if abs(darkrooms.value) == DarkRooms.NONE.value:
            return ""
        if abs(darkrooms.value) == DarkRooms.FEW.value:
            affix = "_drF"
        if abs(darkrooms.value) == DarkRooms.SOME.value:
            affix = "_drS"
        if abs(darkrooms.value) == DarkRooms.MANY.value:
            affix = "_drM"
        if abs(darkrooms.value) == DarkRooms.ALL.value:
            affix = "_drA"
        if darkrooms.value < 0:
            affix += "C"
        return affix

    def getStartingLocation(start_location):
        if start_location.value == StartLocation.SOUTH_CAPE.value:
            return ""
        if start_location.value == StartLocation.SAFE.value:
            return "_sS"
        if start_location.value == StartLocation.UNSAFE.value:
            return "_sU"
        if start_location.value == StartLocation.FORCED_UNSAFE.value:
            return "_sF"

    def getEnemizer(enemizer, boss_shuffle):
        if enemizer.value == Enemizer.NONE.value and not boss_shuffle:
            return ""
        affix = "_en"
        if enemizer.value == Enemizer.BALANCED.value:
            affix += "B"
        if enemizer.value == Enemizer.LIMITED.value:
            affix += "L"
        if enemizer.value == Enemizer.FULL.value:
            affix += "F"
        if enemizer.value == Enemizer.INSANE.value:
            affix += "I"
        if boss_shuffle:
            affix += "-B"
        return affix

    def getFluteOpt(flute):
        if flute.value == FluteOpt.SHUFFLE.value:
            return "-fs"
        if flute.value == FluteOpt.FLUTELESS.value:
            return "-fl"
        return ""

    def getSwitch(switch, param):
        if switch:
            return param
        return ""

    filename = "IoGR" + VERSION
    filename += getDifficulty(settings.difficulty)
    filename += getLogic(settings.logic)
    filename += getGoal(settings.goal, settings.statues, settings.statue_req)
    filename += getStartingLocation(settings.start_location)
    filename += getEnemizer(settings.enemizer, settings.boss_shuffle)
    filename += getEntranceShuffle(settings.coupled_exits, settings.town_shuffle, settings.dungeon_shuffle,
                                   settings.overworld_shuffle)
    filename += getSwitch(settings.orb_rando, "_oX")
    filename += getDarkRooms(settings.darkrooms)
    if (
            settings.open_mode or settings.firebird or settings.ohko or settings.z3 or settings.allow_glitches or settings.flute.value > 0 or settings.infinite_inventory or settings.red_jewel_madness):
        filename += "_v"
        filename += getSwitch(settings.open_mode, "o")
        filename += getSwitch(settings.firebird, "f")
        filename += getSwitch(settings.allow_glitches, "g")
        filename += getSwitch(settings.ohko, "1")
        filename += getSwitch(settings.z3, "z")
        filename += getSwitch(settings.infinite_inventory, "i")
        filename += getFluteOpt(settings.flute)
        filename += getSwitch(settings.red_jewel_madness, "-rjm")
    filename += "_" + str(settings.seed)
    filename += getSwitch(settings.race_mode, "R")
    if extension != "":
        filename += "." + extension

    return filename


class Randomizer:
    statues_required = 0

    def __init__(self, rom_path: str):
        self.rom_path = rom_path
        self.output_folder = os.path.dirname(rom_path) + os.path.sep + "iogr" + os.path.sep

        data_file = open(self.rom_path, "rb")
        self.original_rom_data = data_file.read()
        data_file.close()
        if len(self.original_rom_data) == 0x200200:
            self.original_rom_data = self.original_rom_data[0x200:]  # Strip the 512-byte header
        if len(self.original_rom_data) == 0x200000:
            # Validate a 2MB input
            basehash = hashlib.md5()
            basehash.update(self.original_rom_data)
            if basehash.hexdigest() != 'a7c7a76b4d6f6df389bd631757b91b76':
                raise OffsetError
        elif len(self.original_rom_data) != 0x400000:
            # If caller gives a 4MB input, assume it knows what it's doing; otherwise fail
            raise OffsetError

        log_file_path = os.path.dirname(
            rom_path) + os.path.sep + "iogr" + os.path.sep + "logs" + os.path.sep + "app.log"
        if not os.path.exists(os.path.dirname(log_file_path)):
            os.makedirs(os.path.dirname(log_file_path))

        logging.basicConfig(filename=log_file_path, filemode='w', format='%(message)s', level=logging.DEBUG)
        self.logger = logging.getLogger("IOGR")

    def generate_rom(self, filename: str, settings: RandomizerData, profile_base_filepath=""):
        self.asar_defines = {"DummyRandomizerDefine": "DummyRandomizerDefine"}

        random.seed(settings.seed)
        if settings.race_mode:
            for i in range(random.randint(100, 1000)):
                _ = random.randint(0, 10000)

        statues_required = self.__get_required_statues__(settings)
        statue_req = settings.statue_req.value
        if statue_req == StatueReq.RANDOM_CHOICE.value:
            if random.randint(0, 1):
                statue_req = StatueReq.GAME_CHOICE.value
            else:
                statue_req = StatueReq.PLAYER_CHOICE.value

        ##########################################################################
        #                  Generate hash code for title screen
        ##########################################################################
        hash_str = filename
        h = hashlib.sha256()
        h.update(hash_str.encode())
        hash = h.digest()

        hash_dict = ["/", ".", "[", "]", "*", ",", "+", "-",
                     "2", "3", "4", "5", "6", "7", "8", "9", ":",
                     "(", ")", "?", "A", "B", "C", "D", "E", "F", "G",
                     "H", "I", "J", "K", "L", "M", "N", "P", "Q", "R",
                     "S", "T", "U", "V", "W", "X", "Y", "Z", "'", "<", ">",
                     "#", "a", "b", "c", "d", "e", "f", "g", "h", "i",
                     "j", "k", "m", "n", "o", "p", "q", "r", "s", "t",
                     "u", "v", "w", "x", "y", "z", "{", "}", "=", ";"]

        hash_len = len(hash_dict)

        i = 0
        hash_final = ""
        while i < 6:
            key = hash[i] % hash_len
            hash_final += hash_dict[key]
            i += 1

        self.asar_defines["RandoTitleScreenHashString"] = hash_final

        ##########################################################################
        #                            Randomize Inca tile
        ##########################################################################
        # Set random X/Y for new Inca tile
        inca_x = random.randint(0, 11)
        inca_y = random.randint(0, 5)
        self.asar_defines["IncaTileX"] = inca_x
        self.asar_defines["IncaTileY"] = inca_y

        ##########################################################################
        #                       Randomize heiroglyph order
        ##########################################################################
        # hieroglyph_info = {    # 2-byte text word, sprite addr, tile ID
        #    1: [0xc1c0, 0x81de, 0x84],
        #    2: [0xc3c2, 0x81e4, 0x85],
        #    3: [0xc5c4, 0x81ea, 0x86],
        #    4: [0xc7c6, 0x81f0, 0x8c],
        #    5: [0xc9c8, 0x81f6, 0x8d],
        #    6: [0xcbca, 0x81fc, 0x8e]
        # }
        hieroglyph_order = [1, 2, 3, 4, 5, 6]
        random.shuffle(hieroglyph_order)
        this_pos = 1
        while this_pos < 7:
            this_hiero = hieroglyph_order[this_pos - 1]
            self.asar_defines["HieroOrder" + str(this_pos)] = this_hiero
            # self.asar_defines["HieroSpritePointer"+str(this_pos)] = hieroglyph_info[this_hiero][1]
            # self.asar_defines["HieroItemTile"+str(this_pos)] = hieroglyph_info[this_hiero][2]
            # self.asar_defines["HieroJournalText"+str(this_pos)] = hieroglyph_info[this_hiero][0]
            this_pos += 1

        ##########################################################################
        #                          Randomize Snake Game
        ##########################################################################
        snakes_per_sec = [0.75, 0.85, 1.175, 1.50]         # By level
        if settings.flute.value >= 1:
            snakes_per_sec = [i / 4.0 for i in snakes_per_sec]
        snake_adj = random.uniform(0.9, 1.1)  # Varies snakes per second by +/-10%
        snake_timer = 5 * random.randint(2, 12)  # Timer between 10 and 60 sec (inc 5)
        snake_target = []
        snake_target.append(int(snake_timer * snakes_per_sec[0] * snake_adj))
        snake_target.append(int(snake_timer * snakes_per_sec[1] * snake_adj))
        snake_target.append(int(snake_timer * snakes_per_sec[2] * snake_adj))
        snake_target.append(int(snake_timer * snakes_per_sec[3] * snake_adj))

        snake_frames_str = format((60 * snake_timer) % 256, "02x") + format(int((60 * snake_timer) / 256), "02x")
        snake_target_str = []
        snake_target_str.append(format(int(snake_target[0] / 10), "x") + format(snake_target[0] % 10, "x"))
        snake_target_str.append(format(int(snake_target[1] / 10), "x") + format(snake_target[1] % 10, "x"))
        snake_target_str.append(format(int(snake_target[2] / 10), "x") + format(snake_target[2] % 10, "x"))
        snake_target_str.append(format(int(snake_target[3] / 10), "x") + format(snake_target[3] % 10, "x"))

        # Update snake game logic and text with new values
        self.asar_defines["SnakeGameTimeLimitSeconds"] = snake_timer
        self.asar_defines["SnakeGameTargetEasy"] = int(snake_timer * snakes_per_sec[0] * snake_adj)
        self.asar_defines["SnakeGameTargetIntermediate"] = int(snake_timer * snakes_per_sec[1] * snake_adj)
        self.asar_defines["SnakeGameTargetAdvanced"] = int(snake_timer * snakes_per_sec[2] * snake_adj)
        self.asar_defines["SnakeGameTargetExpert"] = int(snake_timer * snakes_per_sec[3] * snake_adj)
        # The initial space forces the define to resolve as text instead of a number.
        self.asar_defines["SnakeGameTimeLimitSecondsString"] = "_" + str(snake_timer)
        self.asar_defines["SnakeGameTargetEasyString"] = "_" + str(int(snake_timer * snakes_per_sec[0] * snake_adj))
        self.asar_defines["SnakeGameTargetIntermediateString"] = "_" + str(
            int(snake_timer * snakes_per_sec[1] * snake_adj))
        self.asar_defines["SnakeGameTargetAdvancedString"] = "_" + str(int(snake_timer * snakes_per_sec[2] * snake_adj))
        self.asar_defines["SnakeGameTargetExpertString"] = "_" + str(int(snake_timer * snakes_per_sec[3] * snake_adj))

        ##########################################################################
        #                    Randomize Jeweler Reward amounts
        ##########################################################################
        gem = []
        gem_mod = settings.difficulty.value
        if settings.z3:
            gem.append(random.randint(1, 2))
            gem.append(random.randint(3, 4))
            gem.append(random.randint(5, 7))
            gem.append(random.randint(8, 11))
            gem.append(random.randint(12, 17))
            gem.append(random.randint(18, 23))
            gem.append(random.randint(24, 34))
            if settings.goal.value == Goal.RED_JEWEL_HUNT.value:
                gem[6] = [24, 28, 31, 34][gem_mod]
        else:
            gem.append(random.randint(1, 3))
            gem.append(random.randint(4, 6))
            gem.append(random.randint(7, 9))
            gem.append(random.randint(10, 14))
            gem.append(random.randint(16, 21 + gem_mod))
            gem.append(random.randint(23 + gem_mod, 28 + 2*gem_mod))
            gem.append(random.randint(32 + 2*gem_mod, 38 + 4*gem_mod))
            if settings.goal.value == Goal.RED_JEWEL_HUNT.value:
                gem[6] = [35, 40, 45, 50][gem_mod]

        i = 1
        while i <= 7:
            self.asar_defines["Jeweler" + str(i) + "Cost"] = gem[i - 1]
            i += 1

        ##########################################################################
        #                    Randomize Mystic Statue requirement
        ##########################################################################
        statueOrder = [1, 2, 3, 4, 5, 6]
        random.shuffle(statueOrder)
        statues = []
        statues_hex = []

        self.asar_defines["StatuesRequiredCount"] = statues_required
        for i in statueOrder:
            self.asar_defines["Statue" + str(i) + "Required"] = 0

        if statue_req == StatueReq.PLAYER_CHOICE.value:
            self.asar_defines["SettingStatuesPlayerChoice"] = 1
        else:
            self.asar_defines["SettingStatuesPlayerChoice"] = 0
            i = 0
            while i < statues_required:
                if statueOrder[i] == 1:
                    statues.append(1)
                    statues_hex.append(b"\x21")
                    self.asar_defines["Statue1Required"] = 1
                if statueOrder[i] == 2:
                    statues.append(2)
                    statues_hex.append(b"\x22")
                    self.asar_defines["Statue2Required"] = 1
                if statueOrder[i] == 3:
                    statues.append(3)
                    statues_hex.append(b"\x23")
                    self.asar_defines["Statue3Required"] = 1
                if statueOrder[i] == 4:
                    statues.append(4)
                    statues_hex.append(b"\x24")
                    self.asar_defines["Statue4Required"] = 1
                if statueOrder[i] == 5:
                    statues.append(5)
                    statues_hex.append(b"\x25")
                    self.asar_defines["Statue5Required"] = 1
                if statueOrder[i] == 6:
                    statues.append(6)
                    statues_hex.append(b"\x26")
                    self.asar_defines["Statue6Required"] = 1
                i += 1

        ##########################################################################
        #                           Boss shuffle
        ##########################################################################
        # Edge cases:
        # - MQ2 and SA post-defeat lead to the top of Babel
        # - Babel boss and Mansion boss post-defeat lead to Dao (unless the boss is MQ2 or SA)
        # - The top of Babel leads to the dungeon of MQ2
        # - Rama Statues are required by the Mu boss, not by Vamps
        # - S exit from Vamp statue room returns to the dungeon of Vamps
        boss_order = [*range(1, 8)]
        if settings.boss_shuffle:
            non_will_bosses = [5]  # Never forced to play Mummy Queen as Will
            if settings.flute.value == FluteOpt.FLUTELESS.value:
                # Can't beat Castoth as Will without Flute (or generous ability shuffle and lots of patience)
                non_will_bosses.append(1)
            if settings.difficulty.value < 3:  # For non-Extreme seeds:
                boss_order.remove(7)  # - Don't shuffle Solid Arm at all;
                if settings.flute.value == FluteOpt.FLUTELESS.value:
                    non_will_bosses.append(3)  # - Don't require fluteless Vamps.
            if settings.difficulty.value < 2:  # Also, in Easy/Normal, can't be forced to play Vampires as Will
                if 3 not in non_will_bosses:
                    non_will_bosses.append(3)
            random.shuffle(non_will_bosses)

            # Determine statue order, ensuring that non-Will bosses aren't in Will-only rooms
            # (GtWl and Mansion force Will at the boss door, so can't have a non-Will boss)
            for x in non_will_bosses:
                boss_order.remove(x)
            random.shuffle(boss_order)
            non_will_dungeons = [0, 1, 2, 4]  # i.e. dungeons that can have a non-Will boss
            # Assign non_will_bosses to the required number of random non_will_dungeons
            random.shuffle(non_will_dungeons)
            non_will_dungeons = non_will_dungeons[:len(non_will_bosses)]
            non_will_dungeons.sort()
            while non_will_bosses:
                boss = non_will_bosses.pop(0)
                dungeon = non_will_dungeons.pop(0)
                boss_order.insert(dungeon, boss)
            if 7 not in boss_order:
                boss_order.append(7)

            # Patch music headers into new dungeons (beginner and intermediate modes);
            # dungeon 5 (Babel) music is not changed, but is referenced as MinorDungeon by MQ2
            boss_music_card_labels = ["Inca", "SkGn", "Mu", "GtWl", "Pymd", "MinorDungeon", "Mansion"]
            if settings.difficulty.value <= 1:
                i = 0
                while i < 7:
                    if i != 5:
                        boss = boss_order[i]
                        this_dungeon_card = "Map" + boss_music_card_labels[i] + "CardMusic"
                        replacement_card = "DefaultMap" + boss_music_card_labels[boss - 1] + "CardMusic"
                        self.asar_defines[this_dungeon_card] = "!" + replacement_card
                    i += 1
        # Set up assembly defines for boss order
        i = 1
        while i < 8:
            self.asar_defines["Boss" + str(i) + "Id"] = boss_order[i - 1]
            i += 1

        ##########################################################################
        #                   Randomize Location of Kara Portrait
        #       Sets spoiler in Lance's Letter and places portrait sprite
        ##########################################################################
        # Determine random location ID
        kara_location = random.randint(1, 5)
        self.asar_defines["KaraLocation"] = kara_location

        # Set Kara's location and logic mode in RAM switches (for autotracker)
        ## (hmm, I don't think tracker supports this yet?)
        # if settings.logic.value == Logic.COMPLETABLE.value:
        #    logic_int = 0x10 + kara_location
        # elif settings.logic.value == Logic.BEATABLE.value:
        #    logic_int = 0x20 + kara_location
        # else:
        #    logic_int = 0x40 + kara_location
        self.asar_defines["AutotrackerLogicAndKaraVal"] = kara_location  # logic_int

        ##########################################################################
        #                          Have fun with death text
        ##########################################################################
        death_list = []
        death_list.append(
            b"\x2d\x48\xa4\xac\xd6\xa3\xa3\x8e\xac\x87\x80\xa0\xa0\x84\x8d\xa3\xac\xd6\xd7\xcb\xd6\xfe\x85\xa2\x88\x84\x8d\x83\xac\xd7\x73\x88\xa3\xac\xd7\x89\xcb\x4c\x4e\x63\x64\x4b\x69\xac\x83\x84\x80\x83\x2a\x2e\xcb\xac\xac\x6d\x4c\x88\xa2\x80\x82\x8b\x84\xac\x4c\x80\xa8\xc0")
        death_list.append(
            b"\x2d\x69\x8e\xa5\xac\xd7\x9e\x80\xac\x83\x84\x82\x84\x8d\xa4\xac\x85\x84\x8b\x8b\x8e\xa7\x2a\xcb\x48\xac\x87\x80\xa4\x84\xac\xa4\x8e\xac\x8a\x88\x8b\x8b\xac\xa9\x8e\xa5\x2a\x2e\xcb\xac\xac\x6d\x48\x8d\x88\x86\x8e\xac\x4c\x8e\x8d\xa4\x8e\xa9\x80\xc0")
        death_list.append(
            b"\x2d\x64\x8e\xac\x83\x88\x84\xac\xd6\xef\x81\x84\xac\x80\xac\xd6\x95\xcb\x80\x83\xa6\x84\x8d\xa4\xa5\xa2\x84\x2a\x2e\xcb\xac\xac\x6d\x60\x84\xa4\x84\xa2\xac\x41\x80\x8d\x8d\x88\x8d\x86\xc0")
        death_list.append(
            b"\x2d\x4b\x88\x8a\x84\xac\xd6\x94\xa3\x8b\x84\x84\xa0\xac\x82\x8e\x8c\x84\xa3\xcb\x80\x85\xa4\x84\xa2\xac\x87\x80\xa2\x83\xac\xa7\x8e\xa2\x8a\xab\xac\xd6\x94\xa2\x84\xa3\xa4\xcb\x82\x8e\x8c\x84\xa3\xac\x80\x85\xa4\x84\xa2\xac\x80\x8d\xac\x87\x8e\x8d\x84\xa3\xa4\xcb\x8b\x88\x85\x84\x2a\x2e\xac\xac\xac\xac\x6d\x64\xa5\xa2\x81\x8e\xc0")
        death_list.append(
            b"\x2d\x45\x8e\x8b\x8b\x8e\xa7\xac\xd6\xfe\x83\xa2\x84\x80\x8c\xa3\x2a\x2a\x2a\xcb\x83\x84\x80\xa4\x87\xac\x88\xa3\xac\x8d\x8e\xa4\x87\x88\x8d\x86\xab\xac\xd6\xb0\x88\xa3\xcb\x84\xa6\x84\xa2\xa9\xa4\x87\x88\x8d\x86\x2a\x2e\xcb\xac\xac\x6d\x63\xa4\x84\x85\x80\x8d\xac\x4a\x80\xa2\x8b\xac\x63\xa4\x84\x85\x80\x8d\xa3\xa3\x8e\x8d\xc0")
        death_list.append(
            b"\x2d\x40\x8b\x8b\xac\x83\x84\x80\xa4\x87\xa3\xac\x80\xa2\x84\xac\xa3\xa5\x83\x83\x84\x8d\xab\xac\x8d\x8e\xcb\xd6\xb8\x87\x8e\xa7\xac\x86\xa2\x80\x83\xa5\x80\x8b\xac\xa4\x87\x84\xcb\x83\xa9\x88\x8d\x86\xac\x8c\x80\xa9\xac\x81\x84\x2a\x2e\xcb\xac\xac\x6d\x4c\x88\x82\x87\x80\x84\x8b\xac\x4c\x82\x43\x8e\xa7\x84\x8b\x8b\xc0")
        death_list.append(
            b"\x2d\x43\x84\x80\xa4\x87\xac\x88\xa3\xac\xa0\x84\xa2\x87\x80\xa0\xa3\xac\x80\x8d\xcb\x8e\xa2\x83\x84\x80\x8b\xab\xac\x81\xa5\xa4\xac\x88\xa4\xac\x88\xa3\xac\x8d\x8e\xa4\xac\x80\x8d\xcb\x84\xa8\xa0\x88\x80\xa4\x88\x8e\x8d\x2a\x2e\xcb\xac\xac\x6d\x40\x8b\x84\xa8\x80\x8d\x83\xa2\x84\xac\x43\xa5\x8c\x80\xa3\xc0")
        death_list.append(
            b"\x2d\xd6\x62\xd7\x95\x8c\x80\xa4\xa4\x84\xa2\x84\x83\xac\x88\x8d\xcb\x8b\x88\x85\x84\xab\xac\xd6\xf7\x86\x80\xa6\x84\xac\x88\xa4\xcb\xa7\x84\x88\x86\x87\xa4\xab\xac\xa7\x80\xa3\xac\x83\x84\x80\xa4\x87\x2a\x2e\xcb\xac\xac\x6d\x49\x84\x85\x85\xa2\x84\xa9\xac\x44\xa5\x86\x84\x8d\x88\x83\x84\xa3\xc0")
        death_list.append(
            b"\x2d\x4d\x8e\xac\x8e\x8d\x84\xac\xd7\x6d\x8e\xa5\xa4\xac\x8e\x85\xac\xd6\xd6\xcb\xd6\xf5\x80\x8b\x88\xa6\x84\x2a\x2a\x2a\xac\x84\xa8\x82\x84\xa0\xa4\xcb\x80\xa3\xa4\xa2\x8e\x8d\x80\xa5\xa4\xa3\x2a\x2e\xcb\xac\xac\x6d\x63\xa4\x84\xa7\x80\xa2\xa4\xac\x63\xa4\x80\x85\x85\x8e\xa2\x83\xc0")
        death_list.append(
            b"\x2d\x44\xa4\xac\xa4\xa5\xac\x41\xa2\xa5\xa4\x84\x0d\x2e\xcb\xac\xac\x6d\x49\xa5\x8b\x88\xa5\xa3\xac\x42\x80\x84\xa3\x80\xa2\xc0")
        death_list.append(
            b"\x2d\x64\x8e\xac\xa4\x87\x84\xac\xa7\x88\xaa\x80\xa2\x83\xac\x83\x84\x80\xa4\x87\xac\x88\xa3\xcb\x8c\x84\xa2\x84\x8b\xa9\xac\x80\xac\x81\x84\x8b\x88\x84\x85\x2a\x2e\xcb\xac\xac\x6d\x43\x84\x84\xa0\x80\x8a\xac\x42\x87\x8e\xa0\xa2\x80\xc0")
        death_list.append(
            b"\x2d\x44\xa6\x84\xa2\xa9\xac\xd6\xdf\xa9\x8e\xa5\xac\x83\x88\x84\xab\xac\x88\xa4\xcb\x87\xa5\xa2\xa4\xa3\x2a\x2e\xcb\xac\xac\x6d\x49\x8e\xa3\x87\xac\x47\x84\x8d\x83\x84\xa2\xa3\x8e\x8d\xc0")
        death_list.append(
            b"\x2d\x48\x85\xac\x48\xac\xd6\x98\xa4\x8e\xac\x83\x88\x84\xab\xac\x8b\x84\xa4\xac\x8c\x84\xcb\x83\x88\x84\xac\x85\x88\x86\x87\xa4\x88\x8d\x86\x2a\x2e\xcb\xac\xac\x6d\x46\x80\x81\xa2\x88\x84\x8b\xac\x46\x80\xa2\x82\x88\x80\xac\x4c\x80\xa2\xa1\xa5\x84\xaa\xc0")
        death_list.append(
            b"\x2d\x44\x80\x82\x87\xac\x83\x84\x80\xa4\x87\xac\x88\xa3\xac\x80\xa3\xac\xa5\x8d\x88\xa1\xa5\x84\xcb\x80\xa3\xac\x84\x80\x82\x87\xac\x8b\x88\x85\x84\x2a\x2e\xcb\xac\xac\x6d\x4d\x88\x82\x87\x8e\x8b\x80\xa3\xac\x67\x8e\x8b\xa4\x84\xa2\xa3\xa4\x8e\xa2\x85\x85\xc0")
        death_list.append(
            b"\x2d\x48\x8d\xac\xd6\xb0\xa9\x8e\xa5\xac\xd6\x78\xd6\xb5\xcb\x85\x8e\xa2\xa7\x80\xa2\x83\xac\xd6\xab\x81\x80\x82\x8a\x2a\x2e\xcb\xac\xac\x6d\x4c\x80\xa2\x8a\xac\x4c\x80\xa2\xa3\x8b\x80\x8d\x83\xc0")
        death_list.append(
            b"\x2d\x48\xac\x82\x80\x8d\xac\xd6\x91\xa4\x87\x84\xac\x83\x80\x88\xa3\x88\x84\xa3\xcb\x86\xa2\x8e\xa7\x88\x8d\x86\xac\xd6\xbe\x8c\x84\x2a\x2e\xcb\xac\xac\x6d\x49\x8e\x87\x8d\xac\x4a\x84\x80\xa4\xa3\xc0")
        death_list.append(
            b"\x2d\x40\xac\xa0\x84\xa2\xa3\x8e\x8d\x0e\xa3\xac\x83\x84\xa3\xa4\x88\x8d\xa9\xac\x8e\x85\xa4\x84\x8d\xcb\x84\x8d\x83\xa3\xac\xd6\x74\x87\x88\xa3\xac\x83\x84\x80\xa4\x87\x2a\x2e\xcb\xac\xac\x6d\x4c\x88\x8b\x80\x8d\xac\x4a\xa5\x8d\x83\x84\xa2\x80\xc0")
        death_list.append(
            b"\x2d\x43\x84\x80\xa4\x87\xac\x88\xa3\xac\xa4\x87\x84\xac\x83\x84\xa3\xa4\xa2\x8e\xa9\x84\xa2\xcb\x80\x8d\x83\xac\x86\x88\xa6\x84\xa2\xac\x8e\x85\xac\xa3\x84\x8d\xa3\x84\x2a\x2e\xcb\xac\xac\x6d\x40\x8d\x83\xa9\xac\x47\x80\xa2\x86\x8b\x84\xa3\x88\xa3\xc0")
        death_list.append(
            b"\x2d\x41\x84\xac\x82\x80\x8b\x8c\x2a\xac\x46\x8e\x83\xac\x80\xa7\x80\x88\xa4\xa3\xac\xa9\x8e\xa5\xcb\x80\xa4\xac\xa4\x87\x84\xac\x83\x8e\x8e\xa2\x2a\x2e\xcb\xac\xac\x6d\x46\x80\x81\xa2\x88\x84\x8b\xac\x46\x80\xa2\x82\x88\x80\xac\x4c\x80\xa2\xa1\xa5\x84\xaa\xc0")
        death_list.append(
            b"\x2d\x67\x84\xac\xd6\x91\xd7\x88\x80\x8b\x88\xa6\x84\xac\xd6\xf6\xcb\xa7\x84\xac\x80\xa2\x84\xac\x82\x8b\x8e\xa3\x84\xa3\xa4\xac\xa4\x8e\xac\x83\x84\x80\xa4\x87\x2a\x2e\xcb\xac\xac\x6d\x4d\x84\x8d\x88\x80\xac\x42\x80\x8c\xa0\x81\x84\x8b\x8b\xc0")
        death_list.append(
            b"\x43\x84\x80\xa4\x87\xac\xa4\xa7\x88\xa4\x82\x87\x84\xa3\xac\x8c\xa9\xac\x84\x80\xa2\x2f\xcb\x2d\x4b\x88\xa6\x84\xab\x2e\xac\x87\x84\xac\xa3\x80\xa9\xa3\x2a\x2a\x2a\xac\x2d\x48\x0e\x8c\xcb\x82\x8e\x8c\x88\x8d\x86\x2a\x2e\xcb\xac\xac\x6d\x66\x88\xa2\x86\x88\x8b\xc0")
        death_list.append(
            b"\x2d\x47\x84\x80\xa6\x84\x8d\xac\x88\xa3\xac\x80\xac\xd7\x90\xd6\xf4\xcb\x80\x8b\x8b\xac\xa4\x87\x84\xac\x83\x8e\x86\xa3\xac\xa9\x8e\xa5\x0e\xa6\x84\xac\xd7\x5d\xcb\x8b\x8e\xa6\x84\x83\xac\xd6\x79\xa4\x8e\xac\x86\xa2\x84\x84\xa4\xac\xa9\x8e\xa5\x2a\x2e\xcb\xac\xac\x6d\x4e\x8b\x88\xa6\x84\xa2\xac\x46\x80\xa3\xa0\x88\xa2\xa4\xaa\xc0")
        death_list.append(
            b"\x2d\x44\xa6\x84\xa2\xa9\xac\x8c\x80\x8d\xac\x83\x88\x84\xa3\x2a\xac\x4d\x8e\xa4\xcb\x84\xa6\x84\xa2\xa9\xac\x8c\x80\x8d\xac\xd7\x95\x8b\x88\xa6\x84\xa3\x2a\x2e\xcb\xac\xac\x6d\x67\x88\x8b\x8b\x88\x80\x8c\xac\x67\x80\x8b\x8b\x80\x82\x84\xc0")
        death_list.append(
            b"\x2d\x40\x8d\x83\xac\x48\xac\xa7\x88\x8b\x8b\xac\xa3\x87\x8e\xa7\xac\xa4\x87\x80\xa4\xcb\x8d\x8e\xa4\x87\x88\x8d\x86\xac\x82\x80\x8d\xac\x87\x80\xa0\xa0\x84\x8d\xac\x8c\x8e\xa2\x84\xcb\x81\x84\x80\xa5\xa4\x88\x85\xa5\x8b\xac\xa4\x87\x80\x8d\xac\x83\x84\x80\xa4\x87\x2a\x2e\xcb\xac\xac\x6d\x67\x80\x8b\xa4\xac\x67\x87\x88\xa4\x8c\x80\x8d\xc0")
        death_list.append(
            b"\x2d\x46\x84\xa4\xac\x81\xa5\xa3\xa9\xac\x8b\x88\xa6\x88\x8d\x0e\x2b\xac\x8e\xa2\xac\x86\x84\xa4\xcb\x81\xa5\xa3\xa9\xac\x83\xa9\x88\x8d\x0e\x2a\x2e\xcb\xac\xac\x6d\x40\x8d\x83\xa9\xac\x43\xa5\x85\xa2\x84\xa3\x8d\x84\xc0")
        death_list.append(
            b"\x2d\x69\x8e\xa5\x0e\xa2\x84\xac\x8a\x88\x8b\x8b\x88\x8d\x0e\xac\x8c\x84\x2b\xcb\x63\x8c\x80\x8b\x8b\xa3\x4f\x2e\xcb\xac\xac\x6d\x47\x80\x8c\x88\x8b\xa4\x8e\x8d\xac\x60\x8e\xa2\xa4\x84\xa2\xc0")
        death_list.append(
            b"\x2d\x43\x84\x80\xa4\x87\xac\x88\xa3\xac\x89\xa5\xa3\xa4\xac\x80\x8d\x8e\xa4\x87\x84\xa2\xcb\xa0\x80\xa4\x87\x2a\xac\x4E\x8d\x84\xac\xa4\x87\x80\xa4\xac\xa7\x84\xac\x80\x8b\x8b\xac\x8c\xa5\xa3\xa4\xcb\xa4\x80\x8a\x84\x2a\x2e\xcb\xac\xac\x6d\x46\x80\x8d\x83\x80\x8b\x85\xc0")
        death_list.append(
            b"\x2d\x43\x84\x80\xa4\x87\xac\x88\xa3\xac\x8d\x8e\xa4\x87\x88\x8d\x86\x2b\xac\x81\xa5\xa4\xac\xa4\x8e\xcb\x8b\x88\xa6\x84\xac\x83\x84\x85\x84\x80\xa4\x84\x83\xac\x80\x8d\x83\xac\x88\x8d\x86\x8b\x8e\xa2\x6d\xcb\x88\x8e\xa5\xa3\xac\x88\xa3\xac\xa4\x8e\xac\x83\x88\x84\xac\x83\x80\x88\x8b\xa9\x2a\x2e\xcb\xac\xac\x6d\x4D\x80\xa0\x8e\x8b\x84\x8e\x8d\xac\x41\x8e\x8d\x80\xa0\x80\xa2\xa4\x84\xc0")
        death_list.append(
            b"\x2d\x44\xa6\x84\xa2\xa9\xac\xa0\x80\xa2\xa4\x88\x8d\x86\xac\x86\x88\xa6\x84\xa3\xac\x80\xcb\x85\x8e\xa2\x84\xa4\x80\xa3\xa4\x84\xac\x8e\x85\xac\x83\x84\x80\xa4\x87\x2b\xac\x84\xa6\x84\xa2\xa9\xcb\xa2\x84\xa5\x8d\x88\x8e\x8d\xac\x80\xac\x87\x88\x8d\xa4\xac\x8e\x85\xac\xa4\x87\x84\xac\xa2\x84\x6d\xcb\xa3\xa5\xa2\xa2\x84\x82\xa4\x88\x8e\x8d\x2a\x2e\xac\x6d\x63\x82\x87\x8e\xa0\x84\x8d\x87\x80\xa5\x84\xa2\xc0")
        death_list.append(
            b"\x2d\x45\x8e\xa2\xac\x8b\x88\x85\x84\xac\x80\x8d\x83\xac\x83\x84\x80\xa4\x87\xac\x80\xa2\x84\xcb\x8e\x8d\x84\x2b\xac\x84\xa6\x84\x8d\xac\x80\xa3\xac\xa4\x87\x84\xac\xa2\x88\xa6\x84\xa2\xac\x80\x8d\x83\xcb\xa4\x87\x84\xac\xa3\x84\x80\xac\x80\xa2\x84\xac\x8e\x8d\x84\x2a\x2e\xcb\xac\x6d\x4A\x87\x80\x8b\x88\x8b\xac\x46\x88\x81\xa2\x80\x8d\xc0")
        death_list.append(
            b"\x2d\x4B\x88\xa6\x84\xac\xa9\x8e\xa5\xa2\xac\x8b\x88\x85\x84\xac\xa4\x87\x80\xa4\xac\xa4\x87\x84\xcb\x85\x84\x80\xa2\xac\x8e\x85\xac\x83\x84\x80\xa4\x87\xac\x82\x80\x8d\xac\x8d\x84\xa6\x84\xa2\xcb\x84\x8d\xa4\x84\xa2\xac\xa9\x8e\xa5\xa2\xac\x87\x84\x80\xa2\xa4\x2a\x2e\xcb\xac\xac\x6d\x64\x84\x82\xa5\x8c\xa3\x84\x87\xc0")
        death_list.append(
            b"\x2d\x40\x82\x87\x88\x84\xa6\x88\x8d\x86\xac\x8b\x88\x85\x84\xac\x88\xa3\xac\x8d\x8e\xa4\xac\xa4\x87\x84\xcb\x84\xa1\xa5\x88\xa6\x80\x8b\x84\x8d\xa4\xac\x8e\x85\xac\x80\xa6\x8e\x88\x83\x88\x8d\x86\xcb\x83\x84\x80\xa4\x87\x2a\x2e\xcb\xac\xac\x6d\x40\xa9\x8d\xac\x62\x80\x8d\x83\xc0")
        death_list.append(
            b"\x2d\x48\x85\xac\xa7\x84\xac\x8c\xa5\xa3\xa4\xac\x83\x88\x84\x2b\xac\xa7\x84\xac\x83\x88\x84\xcb\x83\x84\x85\x84\x8d\x83\x88\x8d\x86\xac\x8e\xa5\xa2\xac\xa2\x88\x86\x87\xa4\xa3\x2a\x2e\xcb\xac\xac\x6d\x63\x88\xa4\xa4\x88\x8d\x86\xac\x41\xa5\x8b\x8b\xc0")
        death_list.append(
            b"\x2d\x64\x87\x84\xac\x82\x8b\x8e\xa3\x84\xa2\xac\xa7\x84\xac\x82\x8e\x8c\x84\xac\xa4\x8e\xac\xa4\x87\x84\xcb\x8d\x84\x86\x80\xa4\x88\xa6\x84\x2b\xac\xa4\x8e\xac\x83\x84\x80\xa4\x87\x2b\xac\xa4\x87\x84\xcb\x8c\x8e\xa2\x84\xac\xa7\x84\xac\x81\x8b\x8e\xa3\xa3\x8e\x8c\x2a\x2e\xcb\xac\xac\x6d\x4C\x8e\x8d\xa4\x86\x8e\x8c\x84\xa2\xa9\xac\x42\x8b\x88\x85\xa4\xc0")
        death_list.append(
            b"\x2d\x48\x85\xac\xa7\x84\xac\x83\x8e\x8d\x0e\xa4\xac\x8a\x8d\x8e\xa7\xac\x8b\x88\x85\x84\x2b\xcb\x87\x8e\xa7\xac\x82\x80\x8d\xac\xa7\x84\xac\x8a\x8d\x8e\xa7\xac\x83\x84\x80\xa4\x87\x0d\x2e\xcb\xac\xac\x6d\x42\x8e\x8d\x85\xa5\x82\x88\xa5\xa3\xc0")
        death_list.append(
            b"\x2d\x48\xac\x83\x8e\x8d\x0e\xa4\xac\x87\x80\xa6\x84\xac\x8d\x8e\xac\x85\x84\x80\xa2\xac\x8e\x85\xcb\x83\x84\x80\xa4\x87\x2a\xac\x4C\xa9\xac\x8e\x8d\x8b\xa9\xac\x85\x84\x80\xa2\xac\x88\xa3\xcb\x82\x8e\x8c\x88\x8d\x86\xac\x81\x80\x82\x8a\xac\xa2\x84\x88\x8d\x82\x80\xa2\x8d\x80\xa4\x84\x83\x2a\x2e\xcb\xac\xac\x6d\x64\xa5\xa0\x80\x82\xac\x63\x87\x80\x8a\xa5\xa2\xc0")
        death_list.append(
            b"\x2d\x4B\x88\x85\x84\xac\x88\xa4\xa3\x84\x8b\x85\xac\x88\xa3\xac\x81\xa5\xa4\xac\xa4\x87\x84\xcb\xa3\x87\x80\x83\x8e\xa7\xac\x8e\x85\xac\x83\x84\x80\xa4\x87\x2b\xac\x80\x8d\x83\xac\xa3\x8e\xa5\x8b\xa3\xcb\x83\x84\xa0\x80\xa2\xa4\x84\x83\xac\x81\xa5\xa4\xac\xa4\x87\x84\xac\xa3\x87\x80\x83\x8e\xa7\xa3\xcb\x8e\x85\xac\xa4\x87\x84\xac\x8b\x88\xa6\x88\x8d\x86\x2a\x2e\xac\x6d\x64\x2a\xac\x41\xa2\x8e\xa7\x8d\x84\xc0")
        death_list.append(
            b"\x2d\x63\x8e\x8c\x84\x8e\x8d\x84\xac\x87\x80\xa3\xac\xa4\x8e\xac\x83\x88\x84\xac\x88\x8d\xcb\x8e\xa2\x83\x84\xa2\xac\xa4\x87\x80\xa4\xac\xa4\x87\x84\xac\xa2\x84\xa3\xa4\xac\x8e\x85\xac\xa5\xa3\xcb\xa3\x87\x8e\xa5\x8b\x83\xac\xa6\x80\x8b\xa5\x84\xac\x8b\x88\x85\x84\xac\x8c\x8e\xa2\x84\x2a\x2e\xcb\xac\xac\x6d\x66\x88\xa2\x86\x88\x8d\x88\x80\xac\x67\x8e\x8e\x8b\x85\xc0")
        death_list.append(
            b"\x2d\x4E\xa5\xa2\xac\x8b\x88\x85\x84\xac\x83\xa2\x84\x80\x8c\xa3\xac\xa4\x87\x84\xcb\x65\xa4\x8e\xa0\x88\x80\x2a\xac\x4E\xa5\xa2\xac\x83\x84\x80\xa4\x87\xac\x80\x82\x87\x88\x84\xa6\x84\xa3\xcb\xa4\x87\x84\xac\x48\x83\x84\x80\x8b\x2a\x2e\xcb\xac\xac\x6d\x66\x88\x82\xa4\x8e\xa2\xac\x47\xa5\x86\x8e\xc0")
        death_list.append(
            b"\x2d\x48\x85\xac\xa9\x8e\xa5\xac\x83\x88\x84\xac\x88\x8d\xac\x80\x8d\xcb\x84\x8b\x84\xa6\x80\xa4\x8e\xa2\x2b\xac\x81\x84\xac\xa3\xa5\xa2\x84\xac\xa4\x8e\xac\xa0\xa5\xa3\x87\xcb\xa4\x87\x84\xac\x65\xa0\xac\x81\xa5\xa4\xa4\x8e\x8d\x2a\x2e\xcb\xac\xac\x6d\x63\x80\x8c\xac\x4B\x84\xa6\x84\x8d\xa3\x8e\x8d\xc0")
        death_list.append(
            b"\x2d\x43\x84\x80\xa4\x87\xac\x88\xa3\xac\xa6\x84\xa2\xa9\xac\x8e\x85\xa4\x84\x8d\xcb\xa2\x84\x85\x84\xa2\xa2\x84\x83\xac\xa4\x8e\xac\x80\xa3\xac\x80\xac\x86\x8e\x8e\x83\xcb\x82\x80\xa2\x84\x84\xa2\xac\x8c\x8e\xa6\x84\x2a\x2e\xcb\xac\xac\x6d\x41\xa5\x83\x83\xa9\xac\x47\x8e\x8b\x8b\xa9\xc0")
        death_list.append(
            b"\x2d\x42\x8e\xa5\xa2\x80\x86\x84\xac\x88\xa3\xac\x81\x84\x88\x8d\x86\xac\xa3\x82\x80\xa2\x84\x83\xcb\xa4\x8e\xac\x83\x84\x80\xa4\x87\x2a\x2a\x2a\xac\x80\x8d\x83\xac\xa3\x80\x83\x83\x8b\x88\x8d\x86\xcb\xa5\xa0\xac\x80\x8d\xa9\xa7\x80\xa9\x2a\x2e\xcb\xac\xac\x6d\x49\x8e\x87\x8d\xac\x67\x80\xa9\x8d\x84\xc0")
        death_list.append(
            b"\x2d\x48\x8d\x80\x82\xa4\x88\xa6\x88\xa4\xa9\xac\x88\xa3\xac\x83\x84\x80\xa4\x87\x2a\x2e\xcb\xac\xac\x6d\x41\x84\x8d\x88\xa4\x8e\xac\x4C\xa5\xa3\xa3\x8e\x8b\x88\x8d\x88\xc0")
        death_list.append(
            b"\x2d\x43\x84\x80\xa4\x87\xac\x88\xa3\xac\xa4\x87\x84\xac\x86\x8e\x8b\x83\x84\x8d\xac\x8a\x84\xa9\xcb\xa4\x87\x80\xa4\xac\x8e\xa0\x84\x8d\xa3\xac\xa4\x87\x84\xac\xa0\x80\x8b\x80\x82\x84\xac\x8e\x85\xcb\x84\xa4\x84\xa2\x8d\x88\xa4\xa9\x2a\x2e\xcb\xac\xac\x6d\x49\x8e\x87\x8d\xac\x4C\x88\x8b\xa4\x8e\x8d\xc0")
        death_list.append(
            b"\x2d\x48\xac\x80\x8b\xa7\x80\xa9\xa3\xac\xa3\x80\xa9\x2b\xac\x82\x8e\x8c\xa0\x8b\x80\x82\x84\x8d\x82\xa9\xcb\x88\xa3\xac\xa4\x87\x84\xac\x8a\x88\xa3\xa3\xac\x8e\x85\xac\x83\x84\x80\xa4\x87\x2a\x2e\xcb\xac\xac\x6d\x63\x87\x80\xa2\x88\xac\x62\x84\x83\xa3\xa4\x8e\x8d\x84\xc0")
        death_list.append(
            b"\x2d\x48\x8d\xac\xa4\x87\x84\xac\x8b\x8e\x8d\x86\xac\xa2\xa5\x8d\xac\xa7\x84\xac\x80\xa2\x84\xcb\x80\x8b\x8b\xac\x83\x84\x80\x83\x2a\x2e\xcb\xac\xac\x6d\x49\x8e\x87\x8d\xac\x4C\x80\xa9\x8d\x80\xa2\x83\xac\x4A\x84\xa9\x8d\x84\xa3\xc0")
        death_list.append(
            b"\x2d\x48\x0e\x8c\xac\x8d\x8e\xa4\xac\x80\x85\xa2\x80\x88\x83\xac\x8e\x85\xac\x83\x84\x80\xa4\x87\x2b\xcb\x81\xa5\xa4\xac\x48\x0e\x8c\xac\x88\x8d\xac\x8d\x8e\xac\x87\xa5\xa2\xa2\xa9\xac\xa4\x8e\xcb\x83\x88\x84\x2a\xac\x48\xac\x87\x80\xa6\x84\xac\xa3\x8e\xac\x8c\xa5\x82\x87\xac\x48\xac\xa7\x80\x8d\xa4\xcb\xa4\x8e\xac\x83\x8e\xac\x85\x88\xa2\xa3\xa4\x2a\x2e\xac\x6d\x63\x2a\xac\x47\x80\xa7\x8a\x88\x8d\x86\xc0")
        death_list.append(
            b"\x2d\x45\x8e\xa2\xac\xa4\x88\xa3\xac\x8d\x8e\xa4\xac\x88\x8d\xac\x8c\x84\xa2\x84\xac\x83\x84\x80\xa4\x87\xcb\xa4\x87\x80\xa4\xac\x8c\x84\x8d\xac\x83\x88\x84\xac\x8c\x8e\xa3\xa4\x2a\x2e\xcb\xac\xac\x6d\x44\x8b\x88\xaa\x80\x81\x84\xa4\x87\xac\x41\x80\xa2\xa2\x84\xa4\xa4\xcb\xac\xac\xac\xac\xac\x41\xa2\x8e\xa7\x8d\x88\x8d\x86\xc0")
        death_list.append(
            b"\x2d\x43\x8e\xac\x8d\x8e\xa4\xac\x85\x84\x80\xa2\xac\x83\x84\x80\xa4\x87\xac\xa3\x8e\xac\x8c\xa5\x82\x87\xcb\x81\xa5\xa4\xac\xa2\x80\xa4\x87\x84\xa2\xac\xa4\x87\x84\xac\x88\x8d\x80\x83\x84\xa1\xa5\x80\xa4\x84\xcb\x8b\x88\x85\x84\x2a\x2e\xcb\xac\xac\x6d\x41\x84\xa2\xa4\x8e\x8b\xa4\xac\x41\xa2\x84\x82\x87\xa4\xc0")
        death_list.append(
            b"\x2d\x4D\x8e\xa4\x87\x88\x8d\x86\xac\x88\x8d\xac\x8b\x88\x85\x84\xac\x88\xa3\xcb\xa0\xa2\x8e\x8c\x88\xa3\x84\x83\xac\x84\xa8\x82\x84\xa0\xa4\xac\x83\x84\x80\xa4\x87\x2a\x2e\xcb\xac\xac\x6d\x4A\x80\x8d\xa9\x84\xac\x67\x84\xa3\xa4\xc0")
        death_list.append(
            b"\x2d\x4C\xa9\xac\x85\x84\x80\xa2\xac\xa7\x80\xa3\xac\x8d\x8e\xa4\xac\x8e\x85\xac\x83\x84\x80\xa4\x87\xcb\x88\xa4\xa3\x84\x8b\x85\x2b\xac\x81\xa5\xa4\xac\x80\xac\x83\x84\x80\xa4\x87\xac\xa7\x88\xa4\x87\x6d\xcb\x8e\xa5\xa4\xac\x8c\x84\x80\x8d\x88\x8d\x86\x2a\x2e\xcb\xac\xac\x6d\x47\xa5\x84\xa9\xac\x4D\x84\xa7\xa4\x8e\x8d\xc0")
        death_list.append(
            b"\x2d\x43\x84\x80\xa4\x87\xac\x88\xa3\xac\x89\xa5\xa3\xa4\xac\x8b\x88\x85\x84\x0e\xa3\xac\x8d\x84\xa8\xa4\xcb\x81\x88\x86\xac\x80\x83\xa6\x84\x8d\xa4\xa5\xa2\x84\x2a\x2e\xcb\xac\xac\x6d\x49\x2a\xac\x4A\x2a\xac\x62\x8e\xa7\x8b\x88\x8d\x86\xc0")
        death_list.append(
            b"\x2d\x41\x84\x82\x80\xa5\xa3\x84\xac\x8e\x85\xac\x88\x8d\x83\x88\x85\x85\x84\xa2\x84\x8d\x82\x84\x2b\xcb\x8e\x8d\x84\xac\x83\x88\x84\xa3\xac\x81\x84\x85\x8e\xa2\x84\xac\x8e\x8d\x84\xcb\x80\x82\xa4\xa5\x80\x8b\x8b\xa9\xac\x83\x88\x84\xa3\x2a\x2e\xcb\xac\xac\x6d\x44\x8b\x88\x84\xac\x67\x88\x84\xa3\x84\x8b\xc0")
        death_list.append(
            b"\x2d\x4C\xa5\xa3\xa4\xac\x8d\x8e\xa4\xac\x80\x8b\x8b\xac\xa4\x87\x88\x8d\x86\xa3\xac\x80\xa4\xcb\xa4\x87\x84\xac\x8b\x80\xa3\xa4\xac\x81\x84\xac\xa3\xa7\x80\x8b\x8b\x8e\xa7\x84\x83\xac\xa5\xa0\xcb\x88\x8d\xac\x83\x84\x80\xa4\x87\x0d\x2e\xcb\xac\xac\x6d\x60\x8b\x80\xa4\x8e\xc0")
        death_list.append(
            b"\x2d\x64\x87\x84\xa2\x84\xac\x88\xa3\xac\x8d\x8e\xac\x83\x84\x80\xa4\x87\x2b\xac\x8e\x8d\x8b\xa9\xcb\x80\xac\x82\x87\x80\x8d\x86\x84\xac\x8e\x85\xac\xa7\x8e\xa2\x8b\x83\xa3\x2a\x2e\xcb\xac\xac\x6d\x42\x87\x88\x84\x85\xac\x63\x84\x80\xa4\xa4\x8b\x84\xc0")
        death_list.append(
            b"\x2d\x48\xac\x87\x80\x83\xac\xa3\x84\x84\x8d\xac\x81\x88\xa2\xa4\x87\xac\x80\x8d\x83\xcb\x83\x84\x80\xa4\x87\xac\x81\xa5\xa4\xac\x87\x80\x83\xac\xa4\x87\x8e\xa5\x86\x87\xa4\xac\xa4\x87\x84\xa9\xcb\xa7\x84\xa2\x84\xac\x83\x88\x85\x85\x84\xa2\x84\x8d\xa4\x2a\x2e\xcb\xac\xac\x6d\x64\x2a\xac\x63\x2a\xac\x44\x8b\x88\x8e\xa4\xc0")
        death_list.append(
            b"\x2d\x42\xa5\xa2\x88\x8e\xa3\x88\xa4\xa9\xac\x88\xa3\xac\x8b\x88\x85\x84\x2a\xcb\x40\xa3\xa3\xa5\x8c\xa0\xa4\x88\x8e\x8d\xac\x88\xa3\xac\x83\x84\x80\xa4\x87\x2a\x2e\xcb\xac\xac\x6d\x4C\x80\xa2\x8a\xac\x60\x80\xa2\x8a\x84\xa2\xc0")
        death_list.append(
            b"\x2d\x48\xac\x85\x84\x84\x8b\xac\x8c\x8e\x8d\x8e\xa4\x8e\x8d\xa9\xac\x80\x8d\x83\xac\x83\x84\x80\xa4\x87\xcb\xa4\x8e\xac\x81\x84\xac\x80\x8b\x8c\x8e\xa3\xa4\xac\xa4\x87\x84\xac\xa3\x80\x8c\x84\x2a\x2e\xcb\xac\xac\x6d\x42\x87\x80\xa2\x8b\x8e\xa4\xa4\x84\xac\x41\xa2\x8e\x8d\xa4\x84\xc0")
        death_list.append(
            b"\x2d\x63\x8e\x8c\x84\xac\xa0\x84\x8e\xa0\x8b\x84\xac\x80\xa2\x84\xac\xa3\x8e\xac\x80\x85\xa2\x80\x88\x83\xcb\xa4\x8e\xac\x83\x88\x84\xac\xa4\x87\x80\xa4\xac\xa4\x87\x84\xa9\xac\x8d\x84\xa6\x84\xa2\xcb\x81\x84\x86\x88\x8d\xac\xa4\x8e\xac\x8b\x88\xa6\x84\x2a\x2e\xcb\xac\xac\x6d\x47\x84\x8d\xa2\xa9\xac\x66\x80\x8d\xac\x43\xa9\x8a\x84\xc0")
        death_list.append(
            b"\x2d\x64\x87\x84\xac\x81\x8e\x83\xa9\xac\x83\x88\x84\xa3\x2b\xac\x81\xa5\xa4\xac\xa4\x87\x84\xcb\xa3\xa0\x88\xa2\x88\xa4\xac\xa4\x87\x80\xa4\xac\xa4\xa2\x80\x8d\xa3\x82\x84\x8d\x83\xa3\xac\x88\xa4\xcb\x82\x80\x8d\x8d\x8e\xa4\xac\x81\x84\xac\xa4\x8e\xa5\x82\x87\x84\x83\xac\x81\xa9\xcb\x83\x84\x80\xa4\x87\x2a\x2e\xac\xac\x6d\x62\x80\x8c\x80\x8d\x80\xac\x4C\x80\x87\x80\xa2\xa3\x87\x88\xc0")
        death_list.append(
            b"\x2d\x67\x87\x84\xa4\x87\x84\xa2\xac\x85\x8e\xa2\xac\x8b\x88\x85\x84\xac\x8e\xa2\xcb\x83\x84\x80\xa4\x87\x2b\xac\x83\x8e\xac\xa9\x8e\xa5\xa2\xac\x8e\xa7\x8d\xac\xa7\x8e\xa2\x8a\xcb\xa7\x84\x8b\x8b\x2a\x2e\xcb\xac\xac\x6d\x49\x8e\x87\x8d\xac\x62\xa5\xa3\x8a\x88\x8d\xc0")
        death_list.append(
            b"\x2d\x64\x87\x84\xac\xa2\x84\xa0\x8e\xa2\xa4\xa3\xac\x8e\x85\xac\x8c\xa9\xac\x83\x84\x80\xa4\x87\xcb\x87\x80\xa6\x84\xac\x81\x84\x84\x8d\xac\x86\xa2\x84\x80\xa4\x8b\xa9\xcb\x84\xa8\x80\x86\x86\x84\xa2\x80\xa4\x84\x83\x2a\x2e\xcb\xac\xac\x6d\x4C\x80\xa2\x8a\xac\x64\xa7\x80\x88\x8d\xc0")
        death_list.append(
            b"\x2d\x64\x87\x84\xac\xa6\x80\x8b\x88\x80\x8d\xa4\xac\x8d\x84\xa6\x84\xa2\xac\xa4\x80\xa3\xa4\x84\xcb\x8e\x85\xac\x83\x84\x80\xa4\x87\xac\x81\xa5\xa4\xac\x8e\x8d\x82\x84\x2a\x2e\xcb\xac\xac\x6d\x67\x88\x8b\x8b\x88\x80\x8c\xac\x63\x87\x80\x8a\x84\xa3\xa0\x84\x80\xa2\x84\xc0")
        death_list.append(
            b"\x2d\x67\x84\xac\x8a\x8d\x8e\xa7\xac\xa4\x87\x84\xac\xa2\x8e\x80\x83\xac\xa4\x8e\xcb\x85\xa2\x84\x84\x83\x8e\x8c\xac\x87\x80\xa3\xac\x80\x8b\xa7\x80\xa9\xa3\xac\x81\x84\x84\x8d\xcb\xa3\xa4\x80\x8b\x8a\x84\x83\xac\x81\xa9\xac\x83\x84\x80\xa4\x87\x2a\x2e\xcb\xac\x6d\x40\x8d\x86\x84\x8b\x80\xac\x43\x80\xa6\x88\xa3\xc0")
        death_list.append(
            b"\x2d\x43\x84\xa3\xa0\x88\xa3\x84\xac\x8d\x8e\xa4\xac\x83\x84\x80\xa4\x87\x2b\xac\x81\xa5\xa4\xcb\xa7\x84\x8b\x82\x8e\x8c\x84\xac\x88\xa4\x2b\xac\x85\x8e\xa2\xac\x8d\x80\xa4\xa5\xa2\x84\xcb\xa7\x88\x8b\x8b\xa3\xac\x88\xa4\xac\x8b\x88\x8a\x84\xac\x80\x8b\x8b\xac\x84\x8b\xa3\x84\x2a\x2e\xcb\xac\xac\x6d\x4C\x80\xa2\x82\xa5\xa3\xac\x40\xa5\xa2\x84\x8b\x88\xa5\xa3\xc0")
        death_list.append(
            b"\x2d\x43\x84\x80\xa4\x87\xac\x88\xa3\xac\x8d\x8e\xa4\xac\xa4\x87\x84\xac\xa7\x8e\xa2\xa3\xa4\xcb\xa4\x87\x80\xa4\xac\x82\x80\x8d\xac\x87\x80\xa0\xa0\x84\x8d\xac\xa4\x8e\xac\x8c\x84\x8d\x2a\x2e\xcb\xac\xac\x6d\x60\x8b\x80\xa4\x8e\xc0")
        death_list.append(
            b"\x2d\x4B\x88\x85\x84\xac\x8b\x84\xa6\x84\x8b\xa3\xac\x80\x8b\x8b\xac\x8c\x84\x8d\x2a\xcb\x43\x84\x80\xa4\x87\xac\xa2\x84\xa6\x84\x80\x8b\xa3\xac\xa4\x87\x84\xcb\x84\x8c\x88\x8d\x84\x8d\xa4\x2a\x2e\xcb\xac\xac\x6d\x46\x84\x8e\xa2\x86\x84\xac\x41\x84\xa2\x8d\x80\xa2\x83\xac\x63\x87\x80\xa7\xc0")
        death_list.append(
            b"\x2d\x43\x84\x80\xa4\x87\xac\x88\xa3\xac\xa4\x87\x84\xac\xa7\x88\xa3\x87\xac\x8e\x85\xcb\xa3\x8e\x8c\x84\x2b\xac\xa4\x87\x84\xac\xa2\x84\x8b\x88\x84\x85\xac\x8e\x85\xac\x8c\x80\x8d\xa9\x2b\xcb\x80\x8d\x83\xac\xa4\x87\x84\xac\x84\x8d\x83\xac\x8e\x85\xac\x80\x8b\x8b\x2a\x2e\xcb\xac\xac\x6d\x4B\xa5\x82\x88\xa5\xa3\xac\x40\x8d\x8d\x80\x84\xa5\xa3\xac\x63\x84\x8d\x84\x82\x80\xc0")
        death_list.append(
            b"\x2d\x43\x84\x80\xa4\x87\xac\xa7\x88\x8b\x8b\xac\x81\x84\xac\x80\xac\x86\xa2\x84\x80\xa4\xcb\xa2\x84\x8b\x88\x84\x85\x2a\xac\x4D\x8e\xac\x8c\x8e\xa2\x84\xcb\x88\x8d\xa4\x84\xa2\xa6\x88\x84\xa7\xa3\x2a\x2e\xcb\xac\xac\x6d\x4A\x80\xa4\x87\x80\xa2\x88\x8d\x84\xac\x47\x84\xa0\x81\xa5\xa2\x8d\xc0\xc0")
        if settings.difficulty.value >= 3:
            death_list.append(
                b"\x2d\x46\x88\xa4\xac\x86\xa5\x83\xac\xa3\x82\xa2\xa5\x81\x2a\x2e\xcb\xac\xac\x6d\x41\x80\x86\xa5\xc0")

        # Will death text
        death_str = death_list[random.randint(0, len(death_list) - 1)]
        self.asar_defines["PlayerDeathText"] = ""
        i = 0
        while i < len(death_str):
            self.asar_defines["PlayerDeathText"] += "$" + format(death_str[i], "02x")
            i += 1
            if i < len(death_str):
                self.asar_defines["PlayerDeathText"] += ","

        ##########################################################################
        #                   Randomize item and ability placement
        ##########################################################################
        done = False
        self.seed_adj = 0
        while not done:
            if self.seed_adj > MAX_RANDO_RETRIES:
                self.logger.error("ERROR: Max number of seed adjustments exceeded")
                raise RecursionError
            elif self.seed_adj > 0:
                if settings.printlevel.value > -1:
                    print("Trying again... attempt", self.seed_adj + 1)
            self.w = World(settings, statues_required, statues, statue_req, kara_location, gem,
                           [inca_x + 1, inca_y + 1], hieroglyph_order, boss_order)
            done = self.w.randomize(self.seed_adj, settings.printlevel.value, settings.break_on_error,
                                    settings.break_on_init)
            if profile_base_filepath != "":
                val_messages = self.w.validate()
                f = open(profile_base_filepath + "_" + format(self.seed_adj, "02") + ".txt", "w")
                f.write(generate_filename(settings, ""))
                f.write("\n\n")
                f.write("Error log:\n")
                if not self.w.errorlog:
                    f.write("No errors\n")
                else:
                    for m in self.w.errorlog:
                        f.write(m + "\n")
                f.write("\n\n")
                f.write("Validation log:\n")
                for m in val_messages:
                    f.write(m + "\n")
                f.write("\n\n")
                f.write("World graph:\n")
                for n in sorted(self.w.graph):
                    f.write(str(n) + ": " + str(self.w.graph[n]) + "\n")
                f.write("\n\n")
                f.write("World edges:\n")
                for e in sorted(self.w.logic):
                    f.write(str(e) + ": " + str(self.w.logic[e]) + "\n")
                f.write("\n\n")
                f.write("World exits:\n")
                for x in sorted(self.w.exits):
                    f.write(str(x) + ": " + str(self.w.exits[x]) + "\n")
                f.write("\n\n")
                f.close()
            self.seed_adj += 1
        self.w.generate_spoiler(VERSION)
        self.w.populate_asar_defines()
        for wdef in self.w.asar_defines:
            self.asar_defines[wdef] = self.w.asar_defines[wdef]

        ##########################################################################
        #                        Randomize Ishtar puzzle
        ##########################################################################
        used_hair = False
        for room, maxchg in [(1, 10), (2, 7), (3, 4), (4, 9)]:
            minchg = 1 if used_hair else 0
            change1 = random.randint(minchg, maxchg)
            if change1 == 0:
                used_hair = True
                minchg = 1
            self.asar_defines["IshtarRoom" + str(room) + "DifferenceIndex1"] = change1
            if settings.difficulty.value >= 2:
                change2 = change1
                while change2 == change1:
                    change2 = random.randint(minchg, maxchg)
                if change2 == 0:
                    used_hair = True
                    minchg = 1
                self.asar_defines["IshtarRoom" + str(room) + "DifferenceIndex2"] = change2
                if settings.difficulty.value >= 3:
                    change3 = change2
                    while change3 == change2 or change3 == change1:
                        change3 = random.randint(minchg, maxchg)
                    if change3 == 0:
                        used_hair = True
                        minchg = 1
                    self.asar_defines["IshtarRoom" + str(room) + "DifferenceIndex3"] = change3

        ##########################################################################
        #                                   Plugins
        ##########################################################################
        if self.w.goal == "Apocalypse Gaia":
            self.asar_defines["ApocalypseGaia"] = 1
        else:
            self.asar_defines["ApocalypseGaia"] = 0

        ##########################################################################
        #                          Have fun with final text
        ##########################################################################
        superhero_list = []
        superhero_list.append("I must go,]my people need me!")
        superhero_list.append("_____Up, up, and away!")
        superhero_list.append("_______Up and atom!")
        superhero_list.append("__It's clobberin' time!")
        superhero_list.append("_______For Asgard!")
        superhero_list.append("___It's morphin' time!")
        superhero_list.append("____Back in a flash!")
        superhero_list.append("______I am GROOT!")
        superhero_list.append("_______VALHALLA!")
        superhero_list.append("_______Go Joes!")
        superhero_list.append("Wonder Twin powers]activate!")
        superhero_list.append("____Titans together!")
        superhero_list.append("_______HULK SMASH!")
        superhero_list.append("________Flame on!")
        superhero_list.append("________By Crom!")
        superhero_list.append("_______Excelsior!")
        superhero_list.append("Great Caesar's ghost!")
        superhero_list.append("______Odin's beard!")
        superhero_list.append("____I have the power!")
        superhero_list.append("___Avengers assemble!")
        superhero_list.append("To the Bat-pole, Robin!")
        superhero_list.append("________Shazam!")
        superhero_list.append("_____Bite me, fanboy.")
        superhero_list.append("__Hi-yo Silver... away!")
        superhero_list.append("Here I come to]save the day!")
        superhero_list.append("By the hoary hosts]of Hoggoth!")
        superhero_list.append("____Teen Titans, Go!")
        superhero_list.append("_______Cowabunga!")
        superhero_list.append("_______SPOOOOON!!")
        superhero_list.append("There better be bacon]when I get there...")

        # Assign final text box
        rand_idx = random.randint(0, len(superhero_list) - 1)
        self.asar_defines["TextShadowSuperhero"] = superhero_list[rand_idx]

        ##########################################################################
        #            Pass all defines to assembler and return patch
        ##########################################################################
        if len(self.original_rom_data) == 0x200000:
            romdata = copy.deepcopy(self.original_rom_data) + bytearray(0x200000)
        else:  # Caller provided a pre-expanded input
            romdata = copy.deepcopy(self.original_rom_data)

        self.asar_defines["SettingBossShuffle"] = 1 if settings.boss_shuffle else 0
        self.asar_defines["SettingInfiniteInventory"] = 1 if settings.infinite_inventory else 0
        self.asar_defines["SettingEarlyFirebird"] = 1 if settings.firebird else 0
        self.asar_defines["SettingRedJewelHunt"] = 1 if settings.goal.value is Goal.RED_JEWEL_HUNT.value else 0
        self.asar_defines["SettingRedJewelMadness"] = 1 if settings.red_jewel_madness and not settings.infinite_inventory else 0
        self.asar_defines["SettingOpenMode"] = 1 if settings.open_mode else 0
        self.asar_defines["SettingOHKO"] = 1 if settings.ohko else 0
        self.asar_defines["SettingZ3"] = 1 if settings.z3 else 0
        self.asar_defines["SettingFluteOpt"] = settings.flute.value
        self.asar_defines["SettingEnemizer"] = settings.enemizer.value
        self.asar_defines["SettingTownShuffle"] = 1 if settings.town_shuffle else 0
        self.asar_defines["SettingDungeonShuffle"] = 1 if settings.dungeon_shuffle else 0
        self.asar_defines["SettingOrbRando"] = 1 if settings.orb_rando else 0
        self.asar_defines["SettingDarkRoomsLevel"] = abs(settings.darkrooms.value)
        self.asar_defines["SettingDebug"] = 1 if settings.ingame_debug else 0
        self.asar_defines["SettingDsWarp"] = 1 if settings.ds_warp else 0
        if settings.red_jewel_madness:
            self.asar_defines["InitialHp"] = 40
        elif settings.ohko:
            self.asar_defines["InitialHp"] = 1
        elif settings.z3:
            self.asar_defines["InitialHp"] = 6
        else:
            self.asar_defines["InitialHp"] = 8

        self.asar_defines["OptionMuteMusic"] = 0

        for d in self.asar_defines:
            self.asar_defines[d] = str(self.asar_defines[d])  # The library requires defines to be string type.
        if os.name != "nt":
            asar.init(os.path.dirname(__file__) + os.path.sep + "asar-x64.so")
        else:  # Windows
            os.add_dll_directory(os.path.dirname(__file__))
            if sys.maxsize > 2 ** 32:  # 64-bit
                asar.init("asar-x64.dll")
            else:  # 32-bit
                asar.init("asar-x86.dll")
        self.asar_patch_result = asar.patch(os.path.dirname(__file__) + os.path.sep + "iogr.asr", romdata, [], True,
                                            self.asar_defines)

        if self.asar_patch_result[0]:
            return self.asar_patch_result
        else:
            asar_error_list = asar.geterrors()
            return [False, asar_error_list]

    # Returns a list of dicts of the form {'index': int, 'address': int, 'data': list of ints}
    def generate_legacy_patch(self, filename: str, settings: RandomizerData):
        self.generate_rom(filename, settings)
        if not self.asar_patch_result[0]:
            return False
        legacy_patch = []
        legacy_idx = 0
        new_rom_data = self.asar_patch_result[1]
        import ips
        ips_patch = ips.Patch.create(self.original_rom_data, new_rom_data)
        for record in ips_patch.records:
            if record.rle_size <= 0:
                payload = [x for x in record.content]
            else:
                payload = [record.content[0] for i in range(record.rle_size)]
            legacy_patch.append({'index': legacy_idx, 'address': record.offset, 'data': payload})
            legacy_idx += 1
        return json.dumps(legacy_patch)

    def generate_spoiler(self) -> str:
        return json.dumps(self.w.spoiler)

    def generate_def_dump(self) -> str:
        defkeys_sorted = []
        defs = {}
        if self.asar_patch_result[0]:
            defs = {define: val for define, val in asar.getalldefines().items() if
                    define[:7] not in ["Default", "Monster", "PlayerD", "AG_Spr_", "DG_Spr_"]}
            defkeys_sorted = sorted(defs)
        defdump = ""
        for d in defkeys_sorted:
            defdump += "!" + d + " = " + defs[d] + "\n"
        return defdump

    def generate_config_addrs(self) -> str:
        cfgkeys_sorted = []
        config_labels = {}
        if self.asar_patch_result[0]:
            labels = asar.getalllabels()
            config_labels = {label: val for label, val in labels.items() if label[:7] == "Config_"}
            cfgkeys_sorted = sorted(config_labels)
        addrdump = ""
        for d in cfgkeys_sorted:
            addrdump += d + "\t" + str(int(config_labels[d]) & 0x3fffff) + "\n"
        return addrdump

    def __get_required_statues__(self, settings: RandomizerData) -> int:
        if settings.goal.value == Goal.RED_JEWEL_HUNT.value:
            return 0
        if settings.statues.lower() == "random":
            return random.randint(0, 6)
        return int(settings.statues)



from enum import Enum


class DarkRooms(Enum):
    ALLCURSED = -4
    MANYCURSED = -3
    SOMECURSED = -2
    FEWCURSED = -1
    NONE = 0
    FEW = 1
    SOME = 2
    MANY = 3
    ALL = 4


class Difficulty(Enum):
    EASY = 0
    NORMAL = 1
    HARD = 2
    EXTREME = 3


class DungeonShuffle(Enum):
    NONE = 0
    BASIC = 1
    CHAOS = 2
    CLUSTERED = 3


class Enemizer(Enum):
    NONE = 0
    LIMITED = 1
    BALANCED = 2
    FULL = 3
    INSANE = 4


class EntranceShuffle(Enum):
    NONE = 0
    COUPLED = 1
    UNCOUPLED = 2


class FluteOpt(Enum):
    START = 0
    SHUFFLE = 1
    FLUTELESS = 2


class Goal(Enum):
    DARK_GAIA = 0
    RED_JEWEL_HUNT = 1
    APO_GAIA = 2
    RANDOM_GAIA = 3


class Level(Enum):
    BEGINNER = 0
    INTERMEDIATE = 1
    ADVANCED = 2
    EXPERT = 3


class Logic(Enum):
    COMPLETABLE = 0
    BEATABLE = 1
    CHAOS = 2


class OrbRando(Enum):
    NONE = 0
    BASIC = 1
    ORBSANITY = 2


class Sprite(Enum):
    WILL = "will"
    BAGU = "bagu"
    INVISIBLE = "invisible"
    FREET = "freet"
    SOLAR = "solar"
    SYE = "sye"


class StartLocation(Enum):
    SOUTH_CAPE = 0
    SAFE = 1
    UNSAFE = 2
    FORCED_UNSAFE = 3


class StatueReq(Enum):
    GAME_CHOICE = 0
    PLAYER_CHOICE = 1
    RANDOM_CHOICE = 2


class PrintLevel(Enum):
    SILENT = -1
    ERROR = 0
    WARN = 1
    INFO = 2
    VERBOSE = 3



import argparse
import json
import os
import random
import sys

from ..randomizer.iogr_rom import Randomizer
from ..randomizer.iogr_rom import generate_filename
from ..randomizer.models.enums import Difficulty
from ..randomizer.models.enums import Enemizer
from ..randomizer.models.enums import Goal
from ..randomizer.models.enums import Logic
from ..randomizer.models.enums import StartLocation
from ..randomizer.models.randomizer_data import RandomizerData

parser = argparse.ArgumentParser(description="Generate a randomly seeded ROM")
parser.add_argument('-p', '--path', dest="path", type=str, required=True, help="Path to the base ROM file")
parser.add_argument('-s', '--seed', dest="seed", type=int, required=False, default=random.randint(0, 999999999),
                    help="A seed number, must be a valid integer")
parser.add_argument('-d', '--difficulty', dest="difficulty", type=Difficulty, required=False, default=Difficulty.NORMAL,
                    help="Decreases rewards and increases enemy scaling")
parser.add_argument('-g', '--goal', dest="goal", type=Goal, required=False, default=Goal.DARK_GAIA)
parser.add_argument('-l', '--logic', dest="logic", type=Logic, required=False, default=Logic.COMPLETABLE)
parser.add_argument('-e', '--enemizer', dest="enemizer", type=Enemizer, required=False, default=Enemizer.NONE)
parser.add_argument('-z', '--z3', dest="z3", type=bool, required=False, default=False)
parser.add_argument('-r', '--race', dest="race", type=bool, required=False, default=False)
parser.add_argument('--start-pos', dest="start", type=StartLocation, required=False, default=StartLocation.SOUTH_CAPE)
parser.add_argument('--statues', dest="statues", type=str, required=False, default="4")
parser.add_argument('--firebird', dest="firebird", type=bool, required=False, default=False)
parser.add_argument('--allow-glitches', dest="allow_glitches", type=bool, required=False, default=False)
parser.add_argument('--boss-shuffle', dest="boss_shuffle", type=bool, required=False, default=False,
                    help="Randomizes where bosses are placed "
                         "(this will most likely softlock your game, use at own risk)")
parser.add_argument('--overworld-shuffle', dest="overworld_shuffle", type=bool, required=False, default=False)
parser.add_argument('--dungeon-shuffle', dest="dungeon_shuffle", type=bool, required=False, default=False)
parser.add_argument('--open-mode', dest="open_mode", type=bool, required=False, default=False)

modeParser = parser.add_mutually_exclusive_group(required=False)
modeParser.add_argument('--ohko', dest="ohko", action='store_true')
modeParser.add_argument('--red-jewel-madness', dest="red_jewel_madness", action='store_true')

parser.set_defaults(ohko=False)
parser.set_defaults(red_jewel_madness=False)


def main(argv):
    args = parser.parse_args(argv)
    settings = RandomizerData(args.seed, args.difficulty, args.goal, args.logic, args.statues, args.enemizer,
                              args.start, args.firebird, args.ohko, args.red_jewel_madness,
                              args.allow_glitches, args.boss_shuffle, args.open_mode, args.z3, args.overworld_shuffle,
                              args.race, args.dungeon_shuffle)

    rom_filename = generate_filename(settings, "sfc")
    spoiler_filename = generate_filename(settings, "json")

    randomizer = Randomizer(args.path)
    patch = randomizer.generate_rom(rom_filename, settings)
    if not args.race:
        spoiler = randomizer.generate_spoiler()
        write_spoiler(spoiler, spoiler_filename, args.path)

    write_patch(patch, rom_filename, args.path)


def write_spoiler(spoiler, filename, rom_path):
    f = open(os.path.dirname(rom_path) + os.path.sep + filename, "w+")
    f.write(spoiler)
    f.close()
    print("Spoiler created: " + filename)


def sort_patch(val):
    return val['index']


def write_patch(patch, filename, rom_path):
    original = open(rom_path, "rb")
    randomized = open(os.path.dirname(rom_path) + os.path.sep + filename, "wb")
    randomized.write(original.read())

    original.close()
    data = json.loads(patch)
    data.sort(key=sort_patch)

    for k in data:
        address = int(k['address'])
        value = bytes(k['data'])

        randomized.seek(address)
        randomized.write(value)
    randomized.close()
    print("Patch created: " + filename)


def cli():
    return main(sys.argv[1:])


if __name__ == "__main__":
    cli()



import os
import random
import tkinter
import tkinter.filedialog
import tkinter.messagebox

import bsdiff4
import ips

from ..randomizer.errors import RomNotFoundError
from ..randomizer.errors import OffsetError
from ..randomizer.iogr_rom import Randomizer
from ..randomizer.iogr_rom import VERSION
from ..randomizer.iogr_rom import generate_filename
from ..randomizer.models.enums import *
from ..randomizer.models.randomizer_data import RandomizerData


def find_ROM():
    ROM.delete(0, tkinter.END)
    ROM.insert(10, tkinter.filedialog.askopenfilename())


def load_ROM():
    try:
        f = open("rompath.txt")
        return f.read()
    except:
        return ""


def save_ROM(rompath):
    with open("rompath.txt", "w") as f:
        f.write(rompath)


def generate_seed():
    seed.delete(0, tkinter.END)
    seed.insert(10, str(random.randint(0, 99999999)))


def generate_ROM():
    rompath = ROM.get()
    save_ROM(rompath)
    seed_str = seed.get()

    def get_difficulty():
        d = difficulty.get()
        if d == "Easy":
            return Difficulty.EASY
        if d == "Normal":
            return Difficulty.NORMAL
        if d == "Hard":
            return Difficulty.HARD
        if d == "Extreme":
            return Difficulty.EXTREME

    def get_goal():
        g = goal.get()
        if g == "Dark Gaia":
            return Goal.DARK_GAIA
        if g == "Apocalypse Gaia":
            return Goal.APO_GAIA
        if g == "Random Gaia":
            return Goal.RANDOM_GAIA
        if g == "Red Jewel Hunt":
            return Goal.RED_JEWEL_HUNT

    def get_statue_req():
        s = statue_req.get()
        if s == "Player Choice":
            return StatueReq.PLAYER_CHOICE
        if s == "Random Choice":
            return StatueReq.RANDOM_CHOICE
        else:
            return StatueReq.GAME_CHOICE

    def get_logic():
        l = logic.get()
        if l == "Completable":
            return Logic.COMPLETABLE
        if l == "Beatable":
            return Logic.BEATABLE
        if l == "Chaos":
            return Logic.CHAOS

    def get_flute_opt():
        f = flute_opt.get()
        if f == "Shuffle Flute":
            return FluteOpt.SHUFFLE
        if f == "Fluteless":
            return FluteOpt.FLUTELESS
        return FluteOpt.START

    def get_enemizer():
        e = enemizer.get()
        if e == "None":
            return Enemizer.NONE
        if e == "Limited":
            return Enemizer.LIMITED
        if e == "Balanced":
            return Enemizer.BALANCED
        if e == "Full":
            return Enemizer.FULL
        if e == "Insane":
            return Enemizer.INSANE

    def get_orb_rando():
        g = orb_rando.get()
        if g == "None":
            return False  # OrbRando.NONE
        if g == "Basic":
            return True  # OrbRando.BASIC
        if g == "Orbsanity":
            return True  # OrbRando.ORBSANITY

    def get_darkrooms():
        l = darkrooms_level.get()
        c = darkrooms_cursed.get()
        if l == "None":
            return DarkRooms.NONE
        if l == "All":
            return DarkRooms.ALL
        if l == "Few":
            if c:
                return DarkRooms.FEWCURSED
            return DarkRooms.FEW
        if l == "Some":
            if c:
                return DarkRooms.SOMECURSED
            return DarkRooms.SOME
        if l == "Many":
            if c:
                return DarkRooms.MANYCURSED
            return DarkRooms.MANY

    def get_start_location():
        g = start.get()
        if g == "South Cape":
            return StartLocation.SOUTH_CAPE
        if g == "Safe":
            return StartLocation.SAFE
        if g == "Unsafe":
            return StartLocation.UNSAFE
        if g == "Forced Unsafe":
            return StartLocation.FORCED_UNSAFE

    def get_sprite():
        sp = sprite.get()
        if sp == "Will":
            return Sprite.WILL
        if sp == "Bagu":
            return Sprite.BAGU
        if sp == "Freet":
            return Sprite.FREET
        if sp == "Invisible":
            return Sprite.INVISIBLE
        if sp == "Solar":
            return Sprite.SOLAR
        if sp == "Sye":
            return Sprite.SYE

    def get_printlevel():
        m = printlevel.get()
        if m == "Error":
            return PrintLevel.ERROR
        if m == "Warn":
            return PrintLevel.WARN
        if m == "Info":
            return PrintLevel.INFO
        if m == "Verbose":
            return PrintLevel.VERBOSE
        return PrintLevel.SILENT

    if not seed_str.isdigit():
        tkinter.messagebox.showinfo("ERROR", "Please enter or generate a valid seed")
        return

    try:
        seed_int = int(seed_str)
        settings = RandomizerData(
            seed=seed_int,
            difficulty=get_difficulty(),
            goal=get_goal(),
            logic=get_logic(),
            statues=statues.get(),
            statue_req=get_statue_req(),
            start_location=get_start_location(),
            enemizer=get_enemizer(),
            firebird=firebird.get(),
            ohko=ohko.get(),
            red_jewel_madness=red_jewel_madness.get(),
            allow_glitches=glitches.get(),
            boss_shuffle=boss_shuffle.get(),
            open_mode=open_mode.get(),
            z3=z3_mode.get(),
            coupled_exits=coupled_exits.get(),
            town_shuffle=town_shuffle.get(),
            dungeon_shuffle=dungeon_shuffle.get(),
            overworld_shuffle=overworld_shuffle.get(),
            race_mode=race_mode_toggle.get(),
            flute=get_flute_opt(),
            sprite=get_sprite(),
            orb_rando=get_orb_rando(),
            darkrooms=get_darkrooms(),
            printlevel=get_printlevel(),
            break_on_error=break_on_error.get(),
            break_on_init=break_on_init.get(),
            ingame_debug=ingame_debug.get(),
            infinite_inventory=infinite_inventory.get(),
            ds_warp=ds_warp.get()
        )

        base_filename = generate_filename(settings, "")
        rom_filename = base_filename + ".sfc"
        defs_filename = base_filename + "_defs.txt"
        cfg_filename = base_filename + "_cfg.tsv"
        ips_smc_filename = base_filename + "_smcpatch.ips"
        ips_sfc_filename = base_filename + "_sfcpatch.ips"
        bsdiff_filename = base_filename + "_bspatch.bsdiff"
        spoiler_filename = base_filename + "_spoiler.json"
        randomizer = Randomizer(rompath)
        if do_tests_toggle.get():
            profiling_path = os.path.dirname(rompath) + os.path.sep + "iogr" + os.path.sep + base_filename + os.path.sep
            if not os.path.exists(os.path.dirname(profiling_path)):
                os.makedirs(os.path.dirname(profiling_path))
            test_number = 1
            while test_number <= 40:
                settings.seed = random.randint(0, 2 ** 32 - 1)
                settings.town_shuffle = (test_number % 2)
                settings.dungeon_shuffle = (test_number % 4) // 2
                if not settings.town_shuffle and not settings.dungeon_shuffle:
                    settings.coupled_exits = 1
                else:
                    settings.coupled_exits = (test_number % 8) // 4
                settings.firebird = (settings.seed & 0b1)
                settings.overworld_shuffle = (settings.seed & 0b10) >> 1
                if random.randint(0, 1):
                    settings.ohko = 0
                    settings.red_jewel_madness = (settings.seed & 0b100) >> 2
                else:
                    settings.ohko = (settings.seed & 0b100) >> 2
                    settings.red_jewel_madness = 0
                settings.allow_glitches = (settings.seed & 0b1000) >> 3
                settings.boss_shuffle = (settings.seed & 0b10000) >> 4
                settings.open_mode = (settings.seed & 0b100000) >> 5
                settings.z3 = (settings.seed & 0b1000000) >> 6
                settings.race_mode = (settings.seed & 0b10000000) >> 7
                settings.ingame_debug = ((settings.seed >> 8) & 0b1)
                settings.infinite_inventory = ((settings.seed >> 8) & 0b10) >> 1
                d = ((settings.seed >> 8) & 0b1100) >> 2
                if d == 0:
                    settings.difficulty = Difficulty.EASY
                elif d == 1:
                    settings.difficulty = Difficulty.NORMAL
                elif d == 2:
                    settings.difficulty = Difficulty.HARD
                else:
                    settings.difficulty = Difficulty.EXTREME
                s = ((settings.seed >> 8) & 0b110000) >> 4
                if s == 0:
                    settings.start_location = StartLocation.SOUTH_CAPE
                elif s == 1:
                    settings.start_location = StartLocation.SAFE
                elif s == 2:
                    settings.start_location = StartLocation.UNSAFE
                else:
                    settings.start_location = StartLocation.FORCED_UNSAFE
                if ((settings.seed >> 8) & 0b1000000) >> 6:
                    settings.orb_rando = True
                else:
                    settings.orb_rando = False
                c = ((settings.seed >> 8) & 0b10000000) >> 7
                e = ((settings.seed >> 16) & 0b111)
                if e == 0:
                    settings.enemizer = Enemizer.LIMITED
                elif e == 1:
                    settings.enemizer = Enemizer.BALANCED
                elif e == 2:
                    settings.enemizer = Enemizer.FULL
                elif e == 3:
                    settings.enemizer = Enemizer.INSANE
                else:
                    settings.enemizer = Enemizer.NONE
                f = ((settings.seed >> 19) & 0b11)
                if f == 0:
                    settings.flute = FluteOpt.SHUFFLE
                elif f == 1:
                    settings.flute = FluteOpt.FLUTELESS
                else:
                    settings.flute = FluteOpt.START
                dr = ((settings.seed >> 21) & 0b111)
                if dr == 0:
                    settings.darkrooms = DarkRooms.FEWCURSED if c else DarkRooms.FEW
                elif dr == 1:
                    settings.darkrooms = DarkRooms.SOMECURSED if c else DarkRooms.SOME
                elif dr == 2:
                    settings.darkrooms = DarkRooms.MANYCURSED if c else DarkRooms.MANY
                elif dr == 3:
                    settings.darkrooms = DarkRooms.ALLCURSED
                else:
                    settings.darkrooms = DarkRooms.NONE
                l = ((settings.seed >> 24) & 0b11)
                if l == 0:
                    settings.logic = Logic.CHAOS
                elif l == 1:
                    settings.logic = Logic.BEATABLE
                else:
                    settings.logic = Logic.COMPLETABLE
                g = ((settings.seed >> 26) & 0b111)
                if g == 0:
                    settings.goal = Goal.RED_JEWEL_HUNT
                else:
                    g -= 1
                    settings.goal = Goal.DARK_GAIA
                    if g < 3:
                        settings.statue_req = StatueReq.PLAYER_CHOICE
                    else:
                        g -= 3
                        settings.statue_req = StatueReq.GAME_CHOICE
                    if g == 0:
                        settings.statues = "2"
                    elif g == 1:
                        settings.statues = "6"
                    else:
                        settings.statues = "4"
                patch = randomizer.generate_rom("", settings, profiling_path + "Test" + format(test_number, "02"))
                test_number += 1
            tkinter.messagebox.showinfo("Success", "Profiling complete; results in ./iogr/" + base_filename + "/")
            pass
        elif do_profile.get():
            random.seed(seed_int)
            import cProfile, pstats, io
            profiling_path = os.path.dirname(rompath) + os.path.sep + "iogr" + os.path.sep + base_filename + os.path.sep
            if not os.path.exists(os.path.dirname(profiling_path)):
                os.makedirs(os.path.dirname(profiling_path))
            profile = cProfile.Profile()
            profile.enable()
            test_number = 1
            while test_number <= 20:
                settings.seed = random.randint(0, 99999999)
                patch = randomizer.generate_rom("", settings, profiling_path + "Test" + format(test_number, "02"))
                test_number += 1
            profile.disable()
            statfile = open(profiling_path + os.path.sep + "profile.txt", "w")
            runstats = pstats.Stats(profile, stream=statfile)
            runstats.strip_dirs()
            runstats.sort_stats('time')
            runstats.print_stats()
            statfile.close()
            tkinter.messagebox.showinfo("Success", "Profiling complete; results in ./iogr/" + base_filename + "/")
        else:
            patch = randomizer.generate_rom(rom_filename, settings)
            if not patch[0]:
                tkinter.messagebox.showerror("Error", "Assembling failed. The first error was:" + str(patch[1][0]))
            else:
                write_patch(patch, rom_filename, rompath)
                if not race_mode_toggle.get():
                    spoiler = randomizer.generate_spoiler()
                    write_text_file(spoiler, spoiler_filename, rompath)
                if gen_patches_toggle.get():
                    def_dump = randomizer.generate_def_dump()
                    cfg_dump = randomizer.generate_config_addrs()
                    write_text_file(def_dump, defs_filename, rompath)
                    write_text_file(cfg_dump, cfg_filename, rompath)
                    ips_sfc = bytes(ips.Patch.create(randomizer.original_rom_data, patch[1]))
                    ips_smc = bytes(
                        ips.Patch.create(bytearray(0x200) + randomizer.original_rom_data, bytearray(0x200) + patch[1]))
                    bspatch = bsdiff4.diff(randomizer.original_rom_data, patch[1])
                    write_bin_file(ips_sfc, ips_sfc_filename, rompath)
                    write_bin_file(ips_smc, ips_smc_filename, rompath)
                    write_bin_file(bspatch, bsdiff_filename, rompath)
                tkinter.messagebox.showinfo("Success!", rom_filename + " has been successfully created!")
    except OffsetError:
        tkinter.messagebox.showerror("ERROR",
                                     "This randomizer is only compatible with the (US) version of Illusion of Gaia")
    except RomNotFoundError:
        tkinter.messagebox.showerror("ERROR", "ROM file does not exist")
    except Exception as e:
        tkinter.messagebox.showerror("ERROR", str(e))


def write_text_file(text, filename, rom_path):
    f = open(os.path.dirname(rom_path) + os.path.sep + filename, "w")
    f.write(text)
    f.close()


def write_bin_file(bin, filename, rom_path):
    f = open(os.path.dirname(rom_path) + os.path.sep + filename, "wb")
    f.write(bin)
    f.close()


def write_patch(patch, filename, rom_path):
    randomized = open(os.path.dirname(rom_path) + os.path.sep + filename, "wb")
    randomized.write(patch[1])
    randomized.close()


def diff_help():
    lines = [
        "Increased difficulty generally places required items deeper in dungeons, so more of the game has to be completed.",
        "",
        "Extreme difficulty additionally removes in-game spoilers, and may require difficult tasks like defeating Solid Arm."]
    tkinter.messagebox.showinfo("Difficulties", "\n".join(lines))


def goal_help():
    lines = ["Goal determines the required conditions to beat the game:", "",
             "[DARK/APOCALYPSE/RANDOM] GAIA:", " - Collect the required Mystic Statues (if any)",
             " - Rescue Kara from her portrait using the Magic Dust",
             " - Acquire and use the Aura to gain Shadow's form",
             " - Talk to Gaia in any Dark Space to face and defeat the Gaia of your choice", "",
             "RED JEWEL HUNT:",
             "Collect the appropriate number of Red Jewels, by difficulty, and turn them into the Jeweler:",
             " - Easy: 35 Red Jewels", " - Normal: 40 Red Jewels", " - Hard/Extreme: 50 Red Jewels"]
    tkinter.messagebox.showinfo("Goal", "\n".join(lines))


def logic_help():
    lines = ["Logic determines how items and abilities are placed:", "",
             "COMPLETABLE:", " - All items and locations are accessible",
             " - Freedan abilities will only show up in dungeons", "",
             "BEATABLE:", " - Some non-essential items may be inaccessible",
             " - Freedan abilities will only show up in dungeons", "",
             "CHAOS:", " - Some non-essential items may be inaccessible", " - Freedan abilities may show up in towns"]
    tkinter.messagebox.showinfo("Logic Modes", "\n".join(lines))


def variant_help():
    lines = ["OHKO: You always have 1 HP and perish in one hit.", "",
             "Red Jewel Madness: Start with 40 HP. Each Red Jewel given to the Jeweler reduces HP by 1.", "",
             "Fluteless: Will has no primary weapon, and can only do damage via abilities or as Freedan/Shadow.", "",
             "Z3 Mode: HP and damage values are adjusted to make combat faster and more dangerous. STR and DEF upgrades double/halve damage.",
             "",
             "Glitches: To beat the game, you may be required to take advantage of bugs or other unintended mechanics.",
             "",
             "Early Firebird: As Shadow, you gain the Firebird attack after using the Crystal Ring and rescuing Kara.",
             "",
             "Open Mode: Intercontinental travel via Lola's Letter, the Teapot, the Memory Melody, the Will, and the Large Roast is accessible from the start of the game.",
             "",
             "Race Seed: A spoiler log is not generated, and the chosen seed is transformed to a new secret seed. Useful for racing if exchanging files isn't possible.",
             ""]
    tkinter.messagebox.showinfo("Variants", "\n".join(lines))


def enemizer_help():
    lines = ["The following enemy shuffle modes are available:", "",
             "LIMITED:", " - Enemies only appear within their own dungeons", "",
             "BALANCED:", " - Enemies can show up in any dungeon, but retain the stats of the enemies they replace", "",
             "FULL:", " - Enemies can show up in any dungeon with their normal stats", "", "INSANE:",
             " - Same as Full, but enemy stats are shuffled"]
    tkinter.messagebox.showinfo("Enemizer", "\n".join(lines))


def start_help():
    lines = ["You can change this setting to affect where you start the game.",
             "Where you start the game is also where Gaia sends you when you warp to start.", "",
             "SOUTH CAPE:", " - You start the game in South Cape, as normal", "",
             "SAFE:", " - You start the game in a random town, in front of a Dark Space", "",
             "UNSAFE:", " - You start the game in front of a random Dark Space, could be in a town or a dungeon", "",
             "FORCED UNSAFE:", " - You're guaranteed to start the game in the middle of a dungeon"]
    tkinter.messagebox.showinfo("Start Location", "\n".join(lines))


def overworld_shuffle_help():
    lines = ["Overworld Shuffle randomizes which overworld-connected rooms are on each continent for overworld travel."]
    tkinter.messagebox.showinfo("Overworld Shuffle", "\n".join(lines))


def entrance_shuffle_help():
    lines = ["Town Shuffle randomizes doors and other transitions outside of dungeons.",
             "Dungeon Shuffle randomizes transitions within dungeons.",
             "",
             "Coupled transitions are reversible: after taking an exit, you can backtrack through the paired exit on the other side.",
             "If coupling is turned off, backtracking generally won't return you to the room you came from."]
    tkinter.messagebox.showinfo("Transition Shuffles", "\n".join(lines))


def orb_rando_help():
    lines = [
        "Shuffles into the item pool the orbs that open doors and barriers, produced when some monsters are destroyed.",
        "Orb-generating monsters grant a random item when they're destroyed."]
    tkinter.messagebox.showinfo("Orb Rando", "\n".join(lines))


def darkrooms_help():
    lines = ["Random dungeon rooms are made dark, generally in clusters on side branches or near the dungeon end.",
             "",
             "The Dark Glasses and the Crystal Ring let you see in darkness. The game won't require you to enter a dark room without both of these items, unless you choose Cursed Darkness.",
             "",
             "If Cursed, you may have to traverse dark rooms without a light source. Darkness may spread out of side branches into central dungeon rooms."]
    tkinter.messagebox.showinfo("Dark Rooms", "\n".join(lines))


def dr_maybe_set_cursed(drlevel):
    if darkrooms_level.get() == "All":
        darkrooms_cursed_checkbox.config(state='disabled')
        darkrooms_cursed.set(True)
    else:
        darkrooms_cursed_checkbox.config(state='normal')


def er_maybe_force_coupled():
    if not dungeon_shuffle.get() and not town_shuffle.get():
        coupled_exits.set(True)
        coupled_exits_checkbox.config(state='disabled')
    else:
        coupled_exits_checkbox.config(state='normal')


def checkbox_clear_rjm():
    red_jewel_madness.set(False)


def on_enable_rjm():
    checkbox_clear_ohko()
    checkbox_clear_infinv()


def checkbox_clear_ohko():
    ohko.set(False)


def checkbox_clear_infinv():
    infinite_inventory.set(False)


def goal_maybe_change_statues(goalchoice):
    if goal.get() == "Red Jewel Hunt":
        statues.set("0")


def statues_maybe_change_goal(statuechoice):
    if statues.get() != "0" and goal.get() == "Red Jewel Hunt":
        goal.set("Dark Gaia")


def checkbox_clear_profile():
    do_profile.set(False)


def checkbox_clear_tests():
    do_tests_toggle.set(False)


root = tkinter.Tk()
root.title("Illusion of Gaia Randomizer (v." + VERSION + ")")
# if os.name == 'nt':
#     root.wm_iconbitmap('iogr.ico')
# else:
#     root.tk.call('wm', 'iconphoto', root._w, tkinter.PhotoImage(file=os.path.join(".",'iogr.png')))

# Add a grid
mainframe = tkinter.Frame(root)
mainframe.grid(column=2, row=9, sticky=(tkinter.N, tkinter.W, tkinter.E, tkinter.S))
mainframe.columnconfigure(0, weight=1)
mainframe.rowconfigure(0, weight=1)
mainframe.pack(pady=20, padx=20)

tkinter.Label(mainframe, text="Base ROM File").grid(row=0, column=0, sticky=tkinter.W)
tkinter.Label(mainframe, text="Seed").grid(row=1, column=0, sticky=tkinter.W)
tkinter.Label(mainframe, text="Difficulty").grid(row=2, column=0, sticky=tkinter.W)
tkinter.Label(mainframe, text="Logic").grid(row=3, column=0, sticky=tkinter.W)
tkinter.Label(mainframe, text="Goal").grid(row=4, column=0, sticky=tkinter.W)
tkinter.Label(mainframe, text="Statues").grid(row=5, column=0, sticky=tkinter.W)
tkinter.Label(mainframe, text="Start Location").grid(row=6, column=0, sticky=tkinter.W)
tkinter.Label(mainframe, text="Gameplay Variants").grid(row=7, column=0, sticky=tkinter.W)
tkinter.Label(mainframe, text="Enemizer").grid(row=9, column=0, sticky=tkinter.W)
tkinter.Label(mainframe, text="Overworld Shuffle").grid(row=16, column=0, sticky=tkinter.W)
tkinter.Label(mainframe, text="Entrance Shuffle").grid(row=17, column=0, sticky=tkinter.W)
# tkinter.Label(mainframe, text="Dungeon Shuffle").grid(row=18, column=0, sticky=tkinter.W)
tkinter.Label(mainframe, text="Orb Rando").grid(row=20, column=0, sticky=tkinter.W)
tkinter.Label(mainframe, text="Dark Rooms").grid(row=22, column=0, sticky=tkinter.W)
tkinter.Label(mainframe, text="Dev Tools").grid(row=50, column=0, sticky=tkinter.W)

difficulty = tkinter.StringVar(root)
diff_choices = ["Easy", "Normal", "Hard", "Extreme"]
difficulty.set("Normal")

goal = tkinter.StringVar(root)
goal_choices = ["Dark Gaia", "Apocalypse Gaia", "Random Gaia", "Red Jewel Hunt"]
goal.set("Dark Gaia")

logic = tkinter.StringVar(root)
logic_choices = ["Completable", "Beatable", "Chaos"]
logic.set("Completable")

firebird = tkinter.BooleanVar(root)
firebird.set(False)

start = tkinter.StringVar(root)
start_choices = ["South Cape", "Safe", "Unsafe", "Forced Unsafe"]
start.set("South Cape")

ohko = tkinter.BooleanVar(root)
ohko.set(False)

red_jewel_madness = tkinter.BooleanVar(root)
red_jewel_madness.set(False)

infinite_inventory = tkinter.BooleanVar(root)
infinite_inventory.set(False)

glitches = tkinter.BooleanVar(root)
glitches.set(False)

boss_shuffle = tkinter.BooleanVar(root)
boss_shuffle.set(False)

open_mode = tkinter.BooleanVar(root)
open_mode.set(False)

z3_mode = tkinter.BooleanVar(root)
z3_mode.set(False)

overworld_shuffle = tkinter.BooleanVar(root)
overworld_shuffle.set(False)

coupled_exits = tkinter.BooleanVar(root)
coupled_exits.set(True)

town_shuffle = tkinter.BooleanVar(root)
town_shuffle.set(False)

dungeon_shuffle = tkinter.BooleanVar(root)
dungeon_shuffle.set(False)

orb_rando = tkinter.StringVar(root)
orb_rando_choices = ["None", "Orbsanity"]  # formerly allowed "Basic" as well
orb_rando.set("None")

darkrooms_level = tkinter.StringVar(root)
darkrooms_level_choices = ["None", "Few", "Some", "Many", "All"]
darkrooms_level.set("None")

darkrooms_cursed = tkinter.BooleanVar(root)
darkrooms_cursed.set(False)

race_mode_toggle = tkinter.BooleanVar(root)
race_mode_toggle.set(False)

flute_opt = tkinter.StringVar(root)
flute_opt_choices = ["Start with Flute", "Shuffle Flute", "Fluteless"]
flute_opt.set("Start with Flute")

enemizer = tkinter.StringVar(root)
enemizer_choices = ["None", "Limited", "Balanced", "Full", "Insane"]
enemizer.set("None")

sprite = tkinter.StringVar(root)
sprite_choices = ["Will"]  # , "Bagu", "Invisible"]#, "Freet", "Solar", "Sye"]
sprite.set("Will")

statues = tkinter.StringVar(root)
statue_choices = ["0", "1", "2", "3", "4", "5", "6", "Random"]
statues.set("4")

statue_req = tkinter.StringVar(root)
statue_req_choices = ["Game Choice", "Player Choice", "Random Choice"]
statue_req.set("Game Choice")

printlevel = tkinter.StringVar(root)
printlevel_choices = ["Silent", "Error", "Warn", "Info", "Verbose"]
printlevel.set("Silent")
break_on_error = tkinter.BooleanVar(root)
break_on_error.set(False)
break_on_init = tkinter.BooleanVar(root)
break_on_init.set(False)
do_profile = tkinter.BooleanVar(root)
do_profile.set(False)
ingame_debug = tkinter.BooleanVar(root)
ingame_debug.set(False)
gen_patches_toggle = tkinter.BooleanVar(root)
gen_patches_toggle.set(False)
do_tests_toggle = tkinter.BooleanVar(root)
do_tests_toggle.set(False)
ds_warp = tkinter.BooleanVar(root)
ds_warp.set(False)

ROM = tkinter.Entry(mainframe, width=40)
ROM.grid(row=0, column=1)
ROM.insert(0, load_ROM())

seed_frame = tkinter.Frame(mainframe, borderwidth=1)
seed_frame.grid(row=1, column=1)
seed_frame.columnconfigure(0, weight=1)
seed_frame.rowconfigure(0, weight=1)
seed = tkinter.Entry(seed_frame)
seed.pack(side='left')
seed.insert(10, str(random.randint(0, 99999999)))

tkinter.OptionMenu(mainframe,
                   difficulty,
                   *diff_choices).grid(row=2, column=1)
tkinter.OptionMenu(mainframe,
                   logic,
                   *logic_choices).grid(row=3, column=1)
tkinter.OptionMenu(mainframe,
                   goal,
                   *goal_choices,
                   command=goal_maybe_change_statues).grid(row=4, column=1)
statues_frame = tkinter.Frame(mainframe)
statues_frame.grid(row=5, column=1)
tkinter.OptionMenu(statues_frame,
                   statues,
                   *statue_choices,
                   command=statues_maybe_change_goal).pack(side='left')
tkinter.OptionMenu(statues_frame,
                   statue_req,
                   *statue_req_choices).pack(side='left')
tkinter.OptionMenu(mainframe,
                   start,
                   *start_choices).grid(row=6, column=1)

variants_frame = tkinter.Frame(mainframe, borderwidth=1, relief='sunken')
variants_frame.grid(row=7, column=1)
variants_frame.columnconfigure(0, weight=1)
variants_frame.rowconfigure(0, weight=1)

tkinter.Label(variants_frame, text="OHKO:").grid(row=0, column=0, sticky=tkinter.E)
tkinter.Checkbutton(variants_frame,
                    variable=ohko,
                    command=checkbox_clear_rjm).grid(row=0, column=1)

tkinter.Label(variants_frame,
              text="Red Jewel Madness:").grid(row=0, column=3, sticky=tkinter.E)
tkinter.Checkbutton(variants_frame,
                    variable=red_jewel_madness,
                    command=on_enable_rjm).grid(row=0, column=4)

tkinter.Label(variants_frame, text="Z3 Mode:").grid(row=1, column=0, sticky=tkinter.E)
tkinter.Checkbutton(variants_frame,
                    variable=z3_mode).grid(row=1, column=1)

tkinter.Label(variants_frame,
              text="Infinite Inventory:").grid(row=1, column=3, sticky=tkinter.E)
tkinter.Checkbutton(variants_frame,
                    variable=infinite_inventory,
                    command=checkbox_clear_rjm).grid(row=1, column=4)

tkinter.Label(variants_frame,
              text="Glitches:").grid(row=2, column=0, sticky=tkinter.E)
tkinter.Checkbutton(variants_frame,
                    variable=glitches).grid(row=2, column=1)

tkinter.Label(variants_frame,
              text="Early Firebird:").grid(row=2, column=3, sticky=tkinter.E)
tkinter.Checkbutton(variants_frame,
                    variable=firebird).grid(row=2, column=4)

tkinter.Label(variants_frame,
              text="Open:").grid(row=3, column=0, sticky=tkinter.E)
tkinter.Checkbutton(variants_frame,
                    variable=open_mode).grid(row=3, column=1)

tkinter.Label(variants_frame,
              text="Race Seed:").grid(row=3, column=3, sticky=tkinter.E)
tkinter.Checkbutton(variants_frame,
                    variable=race_mode_toggle).grid(row=3, column=4)

tkinter.Label(variants_frame,
              text="Flute:").grid(row=4, column=0, sticky=tkinter.E)
tkinter.OptionMenu(variants_frame,
                   flute_opt,
                   *flute_opt_choices).grid(row=4, column=1, columnspan=4, sticky=tkinter.W)

enemy_rando_frame = tkinter.Frame(mainframe, borderwidth=1)
enemy_rando_frame.grid(row=9, column=1)
enemy_rando_frame.columnconfigure(0, weight=1)
enemy_rando_frame.rowconfigure(0, weight=1)

tkinter.OptionMenu(enemy_rando_frame,
                   enemizer,
                   *enemizer_choices).pack(side='left')

tkinter.Label(enemy_rando_frame,
              text="Boss shuffle:").pack(side='left')
tkinter.Checkbutton(enemy_rando_frame, variable=boss_shuffle).pack(
    side='left')

tkinter.Checkbutton(mainframe,
                    variable=overworld_shuffle).grid(row=16, column=1)

er_frame = tkinter.Frame(mainframe, borderwidth=1, relief='sunken')
er_frame.grid(row=17, column=1)
er_frame.columnconfigure(0, weight=1)
er_frame.rowconfigure(0, weight=1)

tkinter.Label(er_frame,
              text="Towns:").grid(row=0, column=0, sticky=tkinter.E)
tkinter.Checkbutton(er_frame,
                    variable=town_shuffle,
                    command=er_maybe_force_coupled).grid(row=0, column=1)

tkinter.Label(er_frame,
              text="Dungeons:").grid(row=1, column=0, sticky=tkinter.E)
tkinter.Checkbutton(er_frame,
                    variable=dungeon_shuffle,
                    command=er_maybe_force_coupled).grid(row=1, column=1)

tkinter.Label(er_frame,
              text="Coupled:").grid(row=0, column=3, rowspan=2, sticky=tkinter.E)
coupled_exits_checkbox = tkinter.Checkbutton(er_frame,
                                             variable=coupled_exits)
coupled_exits_checkbox.grid(row=0, column=4, rowspan=2, sticky=tkinter.W)
coupled_exits_checkbox.config(state='disabled')

tkinter.OptionMenu(mainframe,
                   orb_rando,
                   *orb_rando_choices).grid(row=20, column=1)

darkrooms_frame = tkinter.Frame(mainframe, borderwidth=1)
darkrooms_frame.grid(row=22, column=1)
tkinter.OptionMenu(darkrooms_frame,
                   darkrooms_level,
                   *darkrooms_level_choices,
                   command=dr_maybe_set_cursed).pack(side='left')
tkinter.Label(darkrooms_frame,
              text="Cursed:").pack(side='left')
darkrooms_cursed_checkbox = tkinter.Checkbutton(darkrooms_frame,
                                                variable=darkrooms_cursed)
darkrooms_cursed_checkbox.pack(side='left')

devtools_frame = tkinter.Frame(mainframe, borderwidth=1, relief='sunken')
devtools_frame.grid(row=50, column=1)
devtools_frame.columnconfigure(0, weight=1)
devtools_frame.rowconfigure(0, weight=1)
tkinter.Label(devtools_frame,
              text="Console:").grid(row=0, column=0, sticky=tkinter.E)
tkinter.OptionMenu(devtools_frame,
                   printlevel,
                   *printlevel_choices).grid(row=0, column=1, columnspan=4, sticky=tkinter.W)

tkinter.Label(devtools_frame,
              text="Break on Error:").grid(row=1, column=0, sticky=tkinter.E)
tkinter.Checkbutton(devtools_frame,
                    variable=break_on_error).grid(row=1, column=1)

tkinter.Label(devtools_frame,
              text="Break on Init:").grid(row=1, column=3, sticky=tkinter.E)
tkinter.Checkbutton(devtools_frame,
                    variable=break_on_init).grid(row=1, column=4)

tkinter.Label(devtools_frame,
              text="In-Game Debug:").grid(row=2, column=0, sticky=tkinter.E)
tkinter.Checkbutton(devtools_frame,
                    variable=ingame_debug).grid(row=2, column=1)

tkinter.Label(devtools_frame,
              text="Profile these:").grid(row=2, column=3, sticky=tkinter.E)
tkinter.Checkbutton(devtools_frame,
                    variable=do_profile,
                    command=checkbox_clear_tests).grid(row=2, column=4)

tkinter.Label(devtools_frame,
              text="Create Patches:").grid(row=3, column=0, sticky=tkinter.E)
tkinter.Checkbutton(devtools_frame,
                    variable=gen_patches_toggle).grid(row=3, column=1)

tkinter.Label(devtools_frame,
              text="Profile all:").grid(row=3, column=3, sticky=tkinter.E)
tkinter.Checkbutton(devtools_frame,
                    variable=do_tests_toggle,
                    command=checkbox_clear_profile).grid(row=3, column=4)

tkinter.Label(devtools_frame,
              text="DS Warp:").grid(row=4, column=0, sticky=tkinter.E)
tkinter.Checkbutton(devtools_frame,
                    variable=ds_warp).grid(row=4, column=1)

tkinter.Button(mainframe, text='Browse...', command=find_ROM).grid(row=0, column=2)
tkinter.Button(seed_frame, text='Random Seed', command=generate_seed).pack(side='left')
tkinter.Button(mainframe, text='Generate ROM', command=generate_ROM).grid(row=1, column=2)

tkinter.Button(mainframe, text='?', command=diff_help).grid(row=2, column=2)
tkinter.Button(mainframe, text='?', command=logic_help).grid(row=3, column=2)
tkinter.Button(mainframe, text='?', command=goal_help).grid(row=4, column=2)
tkinter.Button(mainframe, text='?', command=start_help).grid(row=6, column=2)
tkinter.Button(mainframe, text='?', command=variant_help).grid(row=7, column=2)
tkinter.Button(mainframe, text='?', command=enemizer_help).grid(row=9, column=2)
tkinter.Button(mainframe, text='?', command=overworld_shuffle_help).grid(row=16, column=2)
tkinter.Button(mainframe, text='?', command=entrance_shuffle_help).grid(row=17, column=2)
tkinter.Button(mainframe, text='?', command=orb_rando_help).grid(row=20, column=2)
tkinter.Button(mainframe, text='?', command=darkrooms_help).grid(row=22, column=2)


def main():
    root.mainloop()


if __name__ == "__main__":
    main()



import random

from .enums import *


class RandomizerData:
    def __init__(
        self,
        seed: int = random.randint(0, 999999999),
        difficulty: Difficulty = Difficulty.NORMAL,
        goal: Goal = Goal.DARK_GAIA,
        logic: Logic = Logic.COMPLETABLE,
        statues: str = "4",
        statue_req: StatueReq = StatueReq.GAME_CHOICE,
        start_location: StartLocation = StartLocation.SOUTH_CAPE,
        enemizer: Enemizer = Enemizer.NONE,
        coupled_exits: bool = False,
        town_shuffle: bool = False,
        dungeon_shuffle: bool = False,
        overworld_shuffle: bool = False,
        orb_rando: bool = False,
        darkrooms: DarkRooms = DarkRooms.NONE,
        firebird: bool = False,
        ohko: bool = False,
        red_jewel_madness: bool = False,
        allow_glitches: bool = False,
        boss_shuffle: bool = False,
        open_mode: bool = False,
        z3: bool = False,
        race_mode: bool = False,
        flute: FluteOpt = FluteOpt.START,
        sprite: Sprite = Sprite.WILL,
        printlevel: PrintLevel = PrintLevel.SILENT,
        break_on_error: bool = False,
        break_on_init: bool = False,
        ingame_debug: bool = False,
        infinite_inventory: bool = False,
        ds_warp: bool = True
        ):
        self.seed = seed
        self.difficulty = difficulty
        self.start_location = start_location
        self.goal = goal
        self.statue_req = statue_req
        self.logic = logic
        self.statues = statues
        self.enemizer = enemizer
        self.firebird = firebird
        self.ohko = ohko
        self.red_jewel_madness = red_jewel_madness
        self.allow_glitches = allow_glitches
        self.boss_shuffle = boss_shuffle
        self.coupled_exits = coupled_exits
        self.town_shuffle = town_shuffle
        self.overworld_shuffle = overworld_shuffle
        self.dungeon_shuffle = dungeon_shuffle
        self.orb_rando = orb_rando
        self.darkrooms = darkrooms
        self.open_mode = open_mode
        self.z3 = z3
        self.race_mode = race_mode
        self.flute = flute
        self.sprite = sprite
        self.printlevel = printlevel
        self.break_on_error = break_on_error
        self.break_on_init = break_on_init
        self.ingame_debug = ingame_debug
        self.infinite_inventory = infinite_inventory
        self.ds_warp = ds_warp
